# future
from __future__ import annotations

# stdlib
from collections.abc import Callable
from collections.abc import Iterable
import copy
import hashlib
import textwrap
import time
from typing import Any

# third party
from pydantic import Field
from pydantic import field_validator
from rich.progress import Progress
from typing_extensions import Self

# relative
from ...client.api import ServerIdentity
from ...client.client import SyftClient
from ...client.client import SyftClientSessionCache
from ...serde.serializable import serializable
from ...serde.serialize import _serialize
from ...server.credentials import SyftSigningKey
from ...server.credentials import SyftVerifyKey
from ...service.metadata.server_metadata import ServerMetadata
from ...store.linked_obj import LinkedObject
from ...types.datetime import DateTime
from ...types.identity import Identity
from ...types.identity import UserIdentity
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ...types.syft_object import SyftObject
from ...types.syft_object import short_qual_name
from ...types.transforms import TransformContext
from ...types.transforms import rename
from ...types.transforms import transform
from ...types.uid import UID
from ...util import options
from ...util.colors import SURFACE
from ...util.decorators import deprecated
from ...util.markdown import markdown_as_class_with_fields
from ...util.util import full_name_with_qualname
from ..code.user_code import SubmitUserCode
from ..network.network_service import ServerPeer
from ..network.routes import ServerRoute
from ..network.routes import connection_to_route
from ..request.request import Request
from ..request.request import RequestStatus
from ..response import SyftError
from ..response import SyftException
from ..response import SyftInfo
from ..response import SyftNotReady
from ..response import SyftSuccess
from ..user.user import UserView


@serializable(canonical_name="EventAlreadyAddedException", version=1)
class EventAlreadyAddedException(SyftException):
    pass


@transform(ServerMetadata, ServerIdentity)
def metadata_to_server_identity() -> list[Callable]:
    return [rename("id", "server_id"), rename("name", "server_name")]


@serializable()
class ProjectEvent(SyftObject):
    __canonical_name__ = "ProjectEvent"
    __version__ = SYFT_OBJECT_VERSION_1

    __hash_exclude_attrs__ = ["event_hash", "signature"]

    # 1. Creation attrs
    id: UID
    timestamp: DateTime = Field(default_factory=DateTime.now)
    allowed_sub_types: list | None = []
    # 2. Rebase attrs
    project_id: UID | None = None
    seq_no: int | None = None
    prev_event_uid: UID | None = None
    prev_event_hash: str | None = None
    event_hash: str | None = None
    # 3. Signature attrs
    creator_verify_key: SyftVerifyKey | None = None
    signature: bytes | None = None  # dont use in signing

    def __repr_syft_nested__(self) -> tuple[str, str]:
        return (
            short_qual_name(full_name_with_qualname(self)),
            f"{str(self.id)[:4]}...{str(self.id)[-3:]}",
        )

    def _pre_add_update(self, project: Project) -> None:
        pass

    def rebase(self, project: Project) -> Self:
        prev_event = project.events[-1] if project.events else None
        self.project_id = project.id

        if prev_event and prev_event.seq_no is not None:
            self.prev_event_uid = prev_event.id
            self.prev_event_hash = prev_event.event_hash
            self.seq_no = prev_event.seq_no + 1
        else:
            self.prev_event_uid = project.id
            self.prev_event_hash = project.start_hash
            self.seq_no = 1

        return self

    @property
    def valid(self) -> SyftSuccess | SyftError:
        if self.signature is None:
            return SyftError(message="Sign event first")
        try:
            # Recompute hash
            event_hash_bytes, current_hash = create_project_event_hash(self)
            if current_hash != self.event_hash:
                raise Exception(
                    f"Event hash {current_hash} does not match {self.event_hash}"
                )
            if self.creator_verify_key is None:
                return SyftError(message=f"{self}'s creator_verify_key is None")
            self.creator_verify_key.verify_key.verify(event_hash_bytes, self.signature)
            return SyftSuccess(message="Event signature is valid")
        except Exception as e:
            return SyftError(message=f"Failed to validate message. {e}")

    def valid_descendant(
        self, project: Project, prev_event: Self | None
    ) -> SyftSuccess | SyftError:
        valid = self.valid
        if not valid:
            return valid

        if prev_event:
            prev_event_id: UID | None = prev_event.id
            prev_event_hash = prev_event.event_hash
            prev_seq_no = prev_event.seq_no
        else:
            prev_event_id = project.id
            prev_event_hash = project.start_hash
            prev_seq_no = 0

        if self.prev_event_uid != prev_event_id:
            return SyftError(
                message=f"{self} prev_event_uid: {self.prev_event_uid} "
                "does not match {prev_event_id}"
            )

        if self.prev_event_hash != prev_event_hash:
            return SyftError(
                message=f"{self} prev_event_hash: {self.prev_event_hash} "
                "does not match {prev_event_hash}"
            )

        if (
            (prev_seq_no is not None)
            and (self.seq_no is not None)
            and (self.seq_no != prev_seq_no + 1)
        ):
            return SyftError(
                message=f"{self} seq_no: {self.seq_no} "
                "is not subsequent to {prev_seq_no}"
            )

        if self.project_id != project.id:
            return SyftError(
                message=f"{self} project_id: {self.project_id} "
                "does not match {project.id}"
            )

        if hasattr(self, "parent_event_id"):
            parent_event = project.event_id_hashmap[self.parent_event_id]
            if (
                parent_event.allowed_sub_types is not None
                and type(self) not in parent_event.allowed_sub_types
            ):
                return SyftError(
                    message=f"{self} is not a valid subevent" f"for {parent_event}"
                )
        return SyftSuccess(message=f"{self} is valid descendant of {prev_event}")

    def sign(self, signing_key: SyftSigningKey) -> None:
        if self.creator_verify_key != signing_key.verify_key:
            raise Exception(
                f"creator_verify_key has changed from: {self.creator_verify_key} to "
                f"{signing_key.verify_key}"
            )
        # Calculate Hash
        event_hash_bytes, event_hash = create_project_event_hash(self)
        self.event_hash = event_hash

        # Sign Hash
        signed_obj = signing_key.signing_key.sign(event_hash_bytes)
        self.signature = signed_obj._signature

    def publish(self, project: Project) -> SyftSuccess | SyftError:
        try:
            result = project.add_event(self)
            return result
        except EventAlreadyAddedException:  # nosec
            return SyftSuccess(message="Event already added")


class ProjectEventAddObject(ProjectEvent):
    __canonical_name__ = "ProjectEventAddObject"
    __version__ = SYFT_OBJECT_VERSION_1


class ProjectEventAddLink(ProjectEvent):
    __canonical_name__ = "ProjectEventAddLink"
    __version__ = SYFT_OBJECT_VERSION_1


# Project Sub Event are the events which tend to describe the main events
# For example, if a project event is created, then the project sub event will be the
# 1. ProjectThreadMessage
# 2. Emojis
# 3. Approval for request based events
# Mainly these events are used to describe the main events
# For example, we could make our messaging system to have only one level of thread messages
# it would also help us define the expected sub events for each event
# such that only allowed events could be the sub type of the main event
class ProjectSubEvent(ProjectEvent):
    __canonical_name__ = "ProjectSubEvent"
    __version__ = SYFT_OBJECT_VERSION_1

    parent_event_id: UID


@serializable()
class ProjectThreadMessage(ProjectSubEvent):
    __canonical_name__ = "ProjectThreadMessage"
    __version__ = SYFT_OBJECT_VERSION_1

    message: str


@serializable()
class ProjectMessage(ProjectEventAddObject):
    __canonical_name__ = "ProjectMessage"
    __version__ = SYFT_OBJECT_VERSION_1

    message: str
    allowed_sub_types: list[type] = [ProjectThreadMessage]

    def reply(self, message: str) -> ProjectMessage:
        return ProjectThreadMessage(message=message, parent_event_id=self.id)


@serializable()
class ProjectRequestResponse(ProjectSubEvent):
    __canonical_name__ = "ProjectRequestResponse"
    __version__ = SYFT_OBJECT_VERSION_1

    response: bool


@serializable()
class ProjectRequest(ProjectEventAddObject):
    __canonical_name__ = "ProjectRequest"
    __version__ = SYFT_OBJECT_VERSION_1

    linked_request: LinkedObject
    allowed_sub_types: list[type] = [ProjectRequestResponse]

    @field_validator("linked_request", mode="before")
    @classmethod
    def _validate_linked_request(cls, v: Any) -> LinkedObject:
        if isinstance(v, Request):
            linked_request = LinkedObject.from_obj(v, server_uid=v.server_uid)
            return linked_request
        elif isinstance(v, LinkedObject):
            return v
        else:
            raise ValueError(
                f"linked_request should be either Request or LinkedObject, got {type(v)}"
            )

    @property
    def request(self) -> Request:
        return self.linked_request.resolve

    __repr_attrs__ = [
        "request.status",
        "request.changes[-1].code.service_func_name",
    ]

    def _repr_markdown_(self, wrap_as_python: bool = True, indent: int = 0) -> str:
        func_name = None
        if len(self.request.changes) > 0:
            func_name = self.request.changes[-1].code.service_func_name
        repr_dict = {
            "request.status": self.request.status,
            "request.changes[-1].code.service_func_name": func_name,
        }
        return markdown_as_class_with_fields(self, repr_dict)

    def approve(self) -> ProjectRequestResponse:
        result = self.request.approve()
        if isinstance(result, SyftError):
            return result
        return ProjectRequestResponse(response=True, parent_event_id=self.id)

    def accept_by_depositing_result(
        self, result: Any, force: bool = False
    ) -> SyftError | SyftSuccess:
        return self.request.accept_by_depositing_result(result=result, force=force)

    # TODO: To add deny requests, when deny functionality is added

    def status(self, project: Project) -> SyftInfo | SyftError | None:
        """Returns the status of the request.

        Args:
            project (Project): Project object to check the status

        Returns:
            str: Status of the request.

        During Request  status calculation, we do not allow multiple responses
        """
        responses: list[ProjectEvent] = project.get_children(self)
        if len(responses) == 0:
            return SyftInfo(
                "No one has responded to the request yet. Kindly recheck later 🙂"
            )
        elif len(responses) > 1:
            return SyftError(
                message="The Request Contains more than one Response"
                "which is currently not possible"
                "The request should contain only one response"
                "Kindly re-submit a new request"
                "The Syft Team is working on this issue to handle multiple responses"
            )
        response = responses[0]
        if not isinstance(response, ProjectRequestResponse):
            return SyftError(  # type: ignore[unreachable]
                message=f"Response : {type(response)} is not of type ProjectRequestResponse"
            )

        print("Request Status : ", "Approved" if response.response else "Denied")

        return None


def poll_creation_wizard() -> tuple[str, list[str]]:
    w = textwrap.TextWrapper(initial_indent="\t", subsequent_indent="\t")

    welcome_msg = "Welcome to the Poll Creation Wizard 🧙‍♂️ 🪄!!!"

    description1 = """You've arrived here because you were interested in a creating poll.
A poll is a way to gather opinions or information from a group of people.
It typically involves asking a specific question and providing a set of answer choices for respondents to choose from"""

    description2 = """In a poll, the question is the inquiry being asked of the participants.
It should be a multiple-choice. The choices are the options that respondents
are given to select as their answer to the question.For example, a poll question might be
"Which is your favorite color?" with answer choices of red, blue, green, and yellow.
Participants can then select the answer that best represents their opinion or preference"""

    description3 = """Since you didn't pass in questions, choices into .create_poll() (or you did so incorrectly),
this wizard is going to guide you through the process of creating a poll"""

    description4 = """In this wizard, we're going to ask you for a question and list of choices
to create the poll. The Questions and choices are converted to strings"""

    print("\t" + "=" * 69)
    print(w.fill(welcome_msg))
    print("\t" + "=" * 69)
    print()
    print(w.fill(description1))
    print()
    print(w.fill(description2))
    print()
    print(w.fill(description3))
    print()
    print(w.fill(description4))
    print()

    print("\tDo you understand, and are you ready to proceed? (yes/no)")
    print()
    consent = str(input("\t"))
    print()

    if consent == "no":
        raise Exception("User cancelled poll creation wizard!")

    print("\tExcellent! Let's begin!")

    print()

    print("\t" + "-" * 69)
    print()

    print(w.fill("Question 1: Input a question to ask in the poll"))
    print()
    print(w.fill("Examples:"))
    print("\t - What is your favorite type of food?")
    print("\t - What day shall we meet?")
    print("\t - Do you believe that climate change is a serious problem?")
    print()
    print()
    question = input("\t")
    print()

    print("\t" + "-" * 69)
    print()
    print(
        w.fill(
            "Question 2: Enter the number of choices, you would like to have in the poll"
        )
    )
    print()
    while True:
        try:
            num_choices = int(input("\t"))
        except ValueError:
            print()
            print(
                w.fill("Number of choices, should be an integer.Kindly re-enter again.")
            )
            print()
            continue
        break

    print()
    print("\t" + "-" * 69)
    print()
    print(w.fill("Excellent! Let's  input each choice for the input"))
    print()
    choices = []
    for idx in range(num_choices):
        print(w.fill(f"Enter Choice {idx+1}"))
        print()
        choice = str(input("\t"))
        choices.append(choice)
        print()
    print("\t" + "=" * 69)

    print()

    print(
        w.fill("All done! You have successfully completed the Poll Creation Wizard! 🎩")
    )
    return (question, choices)


def poll_answer_wizard(poll: ProjectMultipleChoicePoll) -> int:
    w = textwrap.TextWrapper(initial_indent="\t", subsequent_indent="\t")

    welcome_msg = "Welcome to the Poll Answer Wizard 🧙‍♂️ 🪄!!!"

    description1 = """You've arrived here because you were interested in a answering a poll.
A poll is a way to gather opinions or information from a group of people.
It typically involves asking a specific question and providing a set of answer choices for respondents to choose from"""

    description2 = """In a poll, the question is the inquiry being asked of the participants.
It should be a multiple-choice. The choices are the options that respondents
are given to select as their answer to the question.For example, a poll question might be
"Which is your favorite color?" with answer choices of red, blue, green, and yellow.
Participants can then select the answer that best represents their opinion or preference"""

    description3 = """Since you didn't pass in the choices into .answer_poll() (or you did so incorrectly),
this wizard is going to guide you through the process of answering the poll."""

    print("\t" + "=" * 69)
    print(w.fill(welcome_msg))
    print("\t" + "=" * 69)
    print()
    print(w.fill(description1))
    print()
    print(w.fill(description2))
    print()
    print(w.fill(description3))
    print()

    print("\tDo you understand, and are you ready to proceed? (yes/no)")
    print()
    consent = str(input("\t"))
    print()

    if consent == "no":
        raise Exception("User cancelled poll answer wizard!")

    print("\tExcellent! Let's display the poll question")

    print()

    print("\t" + "-" * 69)
    print()

    print(w.fill(f"Question : {poll.question}"))
    print()
    for idx, choice_i in enumerate(poll.choices):
        print(w.fill(f"{idx+1}. {choice_i}"))
        print()

    print("\t" + "-" * 69)
    print()

    print(w.fill("Kindly enter your choice for the poll"))
    print()
    while True:
        try:
            choice: int = int(input("\t"))
            if choice < 1 or choice > len(poll.choices):
                raise ValueError()
        except ValueError:
            print()
            print(
                w.fill(
                    f"Poll Answer should be a natural number between 1 and {len(poll.choices)}"
                )
            )
            print()
            continue
        break

    print("\t" + "=" * 69)
    print()
    print(
        w.fill("All done! You have successfully completed the Poll Answer Wizard! 🎩")
    )
    print()

    return choice


@serializable()
class AnswerProjectPoll(ProjectSubEvent):
    __canonical_name__ = "AnswerProjectPoll"
    __version__ = SYFT_OBJECT_VERSION_1

    answer: int


@serializable()
class ProjectMultipleChoicePoll(ProjectEventAddObject):
    __canonical_name__ = "ProjectPoll"
    __version__ = SYFT_OBJECT_VERSION_1

    question: str
    choices: list[str]
    allowed_sub_types: list[type] = [AnswerProjectPoll]

    @field_validator("choices")
    @classmethod
    def choices_min_length(cls, v: list[str]) -> list[str]:
        if len(v) < 1:
            raise ValueError("choices must have at least one item")
        return v

    def answer(self, answer: int) -> ProjectMessage:
        return AnswerProjectPoll(answer=answer, parent_event_id=self.id)

    def status(
        self, project: Project, pretty_print: bool = True
    ) -> dict | SyftError | SyftInfo | None:
        """Returns the status of the poll

        Args:
            project (Project): Project object to check the status

        Returns:
            str: Status of the poll

        During Poll calculation, a user would have answered the poll many times
        The status of the poll would be calculated based on the latest answer of the user
        """
        poll_answers = project.get_children(self)
        if len(poll_answers) == 0:
            return SyftInfo(message="No one has answered this poll")

        respondents = {}
        for poll_answer in poll_answers[::-1]:
            if not isinstance(poll_answer, AnswerProjectPoll):
                return SyftError(  # type: ignore[unreachable]
                    message=f"Poll answer: {type(poll_answer)} is not of type AnswerProjectPoll"
                )
            creator_verify_key = poll_answer.creator_verify_key

            # Store only the latest response from the user
            identity = project.get_identity_from_key(creator_verify_key)
            if identity not in respondents:
                respondents[identity] = poll_answer.answer
        if pretty_print:
            for respondent, answer in respondents.items():
                print(f"{str(respondent.verify_key)[0:8]}: {answer}")
            print("\nChoices:\n")
            for idx, choice in enumerate(self.choices):
                print(f"{idx+1}: {choice}")
            return None
        else:
            return respondents


class ConsensusModel:
    pass


@serializable(canonical_name="DemocraticConsensusModel", version=1)
class DemocraticConsensusModel(ConsensusModel):
    threshold: float = 50

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, DemocraticConsensusModel):
            return False
        return self.threshold == value.threshold

    def __hash__(self) -> int:
        return hash(self.threshold)


def add_code_request_to_project(
    project: ProjectSubmit | Project,
    code: SubmitUserCode,
    client: SyftClient | Any,
    reason: str | None = None,
) -> SyftError | SyftSuccess:
    # TODO: fix the mypy issue
    if not isinstance(code, SubmitUserCode):
        return SyftError(  # type: ignore[unreachable]
            message=f"Currently we are only support creating requests for SubmitUserCode: {type(code)}"
        )

    if not isinstance(client, SyftClient):
        return SyftError(message="Client should be a valid SyftClient")

    if reason is None:
        reason = f"Code Request for Project: {project.name} has been submitted by {project.created_by}"

    submitted_req = client.api.services.code.request_code_execution(
        code=code, reason=reason
    )
    if isinstance(submitted_req, SyftError):
        return submitted_req

    request_event = ProjectRequest(linked_request=submitted_req)

    if isinstance(project, ProjectSubmit) and project.bootstrap_events is not None:
        project.bootstrap_events.append(request_event)
    else:
        result = project.add_event(request_event)
        if isinstance(result, SyftError):
            return result

    return SyftSuccess(
        message=f"Code request for '{code.func_name}' successfully added to '{project.name}' Project. "
        f"To see code requests by a client, run `[your_client].code`"
    )


@serializable()
class Project(SyftObject):
    __canonical_name__ = "Project"
    __version__ = SYFT_OBJECT_VERSION_1

    __repr_attrs__ = ["name", "description", "created_by"]
    __attr_unique__ = ["name"]

    # TODO: re-add users, members, leader_server_peer
    __hash_exclude_attrs__ = [
        "user_signing_key",
        "start_hash",
        "users",
        "members",
        "leader_server_peer",
        "event_id_hashmap",
    ]

    id: UID | None = None  # type: ignore[assignment]
    name: str
    description: str | None = None
    members: list[ServerIdentity]
    users: list[UserIdentity] = []
    username: str | None = None
    created_by: str
    start_hash: str | None = None
    # WARNING:  Do not add it to hash keys or print directly
    user_signing_key: SyftSigningKey | None = None

    # Project events
    events: list[ProjectEvent] = []
    event_id_hashmap: dict[UID, ProjectEvent] = {}

    # Project sync
    state_sync_leader: ServerIdentity
    leader_server_peer: ServerPeer | None = None

    # Unused
    consensus_model: ConsensusModel
    project_permissions: set[str]
    # store: Dict[UID, Dict[UID, SyftObject]] = {}
    # permissions: Dict[UID, Dict[UID, Set[str]]] = {}

    def _coll_repr_(self) -> dict:
        return {
            "name": self.name,
            "description": self.description,
            "created by": self.created_by,
            "pending requests": self.pending_requests,
        }

    def _repr_html_(self) -> Any:
        return (
            f"""
            <style>
            .syft-project {{color: {SURFACE[options.color_theme]};}}
            </style>
            """
            + "<div class='syft-project'>"
            + f"<h3>{self.name}</h3>"
            + f"<p>{self.description}</p>"
            + f"<p><strong>Created by: </strong>{self.username} ({self.created_by})</p>"
            + self.requests._repr_html_()
            + "<p>To see a list of projects, use command `&lt;your_client&gt;.projects`</p>"
            + "</div>"
        )

    def _broadcast_event(self, project_event: ProjectEvent) -> SyftSuccess | SyftError:
        leader_client = self.get_leader_client(self.user_signing_key)

        return leader_client.api.services.project.broadcast_event(project_event)

    def get_all_identities(self) -> list[Identity]:
        return [*self.members, *self.users]

    def key_in_project(self, verify_key: SyftVerifyKey) -> bool:
        project_verify_keys = [
            identity.verify_key for identity in self.get_all_identities()
        ]

        return verify_key in project_verify_keys

    def get_identity_from_key(
        self, verify_key: SyftVerifyKey
    ) -> list[ServerIdentity | UserIdentity]:
        identities: list[Identity] = self.get_all_identities()
        for identity in identities:
            if identity.verify_key == verify_key:
                return identity
        return SyftError(message=f"Member with verify key: {verify_key} not found")

    def get_leader_client(self, signing_key: SyftSigningKey) -> SyftClient:
        if self.leader_server_peer is None:
            raise Exception("Leader server peer is not set")

        if signing_key is None:
            raise Exception("Signing key is required to create leader client")

        verify_key = signing_key.verify_key

        leader_client = SyftClientSessionCache.get_client_by_uid_and_verify_key(
            verify_key=verify_key, server_uid=self.leader_server_peer.id
        )

        if leader_client is None:
            leader_client = self.leader_server_peer.client_with_key(signing_key)
            SyftClientSessionCache.add_client_by_uid_and_verify_key(
                verify_key=verify_key,
                server_uid=leader_client.id,
                syft_client=leader_client,
            )

        return leader_client

    def has_permission(self, verify_key: SyftVerifyKey) -> bool:
        # Currently the permission function, initially checks only if the
        # verify key is present in the project
        # Later when the project object evolves, we can add more permission checks

        return self.key_in_project(verify_key)

    def _append_event(
        self, event: ProjectEvent, credentials: SyftSigningKey
    ) -> SyftSuccess | SyftError:
        prev_event = self.events[-1] if self.events else None
        valid = event.valid_descendant(self, prev_event)
        if not valid:
            return valid

        result = self._broadcast_event(event)
        if isinstance(result, SyftError):
            return result
        if isinstance(result, SyftNotReady):
            # If the client if out of sync, sync project updates from leader
            self.sync()
            event = event.rebase(project=self)
            event.sign(credentials)
            # Retrying broadcasting the event to leader
            # recursively call _append_event as due to network latency the event could reach late
            # and other events would be being streamed to the leader
            # This scenario could lead to starvation of server trying to sync with the leader
            # This would be solved in our future leaderless approach
            return self._append_event(event=event, credentials=credentials)

        event_copy = copy.deepcopy(event)
        self.events.append(event_copy)
        self.event_id_hashmap[event.id] = event_copy
        return result

    @property
    def event_ids(self) -> Iterable[UID]:
        return self.event_id_hashmap.keys()

    def add_event(
        self,
        event: ProjectEvent,
        credentials: SyftSigningKey | SyftClient | None = None,
    ) -> SyftSuccess | SyftError:
        if event.id in self.event_ids:
            raise EventAlreadyAddedException(f"Event already added. {event}")

        if credentials is None:
            credentials = self.user_signing_key
        elif isinstance(credentials, SyftClient):
            credentials = credentials.credentials

        if not isinstance(credentials, SyftSigningKey):
            raise Exception(f"Adding an event requires a signing key. {credentials}")

        event.creator_verify_key = credentials.verify_key
        event._pre_add_update(self)
        event = event.rebase(self)
        event.sign(credentials)

        result = self._append_event(event, credentials=credentials)
        return result

    def validate_events(self, debug: bool = False) -> SyftSuccess | SyftError:
        current_hash = self.start_hash

        def valid_str(current_hash: int) -> str:
            return f"Project: {self.id} HEAD<{current_hash}> events are valid"

        if len(self.events) == 0:
            return SyftSuccess(message=valid_str(current_hash))

        last_event = None
        for event in self.events:
            result = event.valid_descendant(self, last_event)
            current_hash = event.event_hash

            if debug:
                icon = "✅" if result else "❌"
                prev_event = last_event if last_event is not None else self
                print(
                    f"{icon} {type(event).__name__}: {event.id} "
                    f"after {type(prev_event).__name__}: {prev_event.id}"
                )

            if not result:
                return result
            last_event = event
        return SyftSuccess(message=valid_str(current_hash))

    def get_children(self, event: ProjectEvent) -> list[ProjectEvent]:
        return self.get_events(parent_event_ids=event.id)

    def get_parent(self, parent_uid: UID) -> ProjectEvent | None:
        parent_event = None
        event_query = self.get_events(ids=parent_uid)
        if len(event_query) == 0:
            return parent_event
        elif len(event_query) == 1:
            return event_query[0]
        else:
            raise Exception(f"More than 1 result for {parent_uid}")

    # TODO: add a another view for the project objects
    # to be able to have a Directed Acyclic Graph view of the graph events
    # this would allow to query the sub events effectively
    def get_events(
        self,
        types: type | list[type] | None = None,
        parent_event_ids: UID | list[UID] | None = None,
        ids: UID | list[UID] | None = None,
    ) -> list[ProjectEvent]:
        if types is None:
            types = []
        if isinstance(types, type):
            types = [types]

        if parent_event_ids is None:
            parent_event_ids = []
        if isinstance(parent_event_ids, UID):
            parent_event_ids = [parent_event_ids]

        if ids is None:
            ids = []
        if isinstance(ids, UID):
            ids = [ids]

        results = []
        for event in self.events:
            type_check = False
            if len(types) == 0 or isinstance(event, tuple(types)):
                type_check = True

            parent_check = False
            if (
                len(parent_event_ids) == 0 and not hasattr(event, "parent_event_id")
            ) or (
                hasattr(event, "parent_event_id")
                and event.parent_event_id in parent_event_ids
            ):
                parent_check = True

            id_check = False
            if len(ids) == 0 or event.id in ids:
                id_check = True

            if type_check and parent_check and id_check:
                results.append(event)
        return results

    def create_code_request(
        self,
        obj: SubmitUserCode,
        client: SyftClient | None = None,
        reason: str | None = None,
    ) -> SyftSuccess | SyftError:
        if client is None:
            leader_client = self.get_leader_client(self.user_signing_key)
            res = add_code_request_to_project(
                project=self,
                code=obj,
                client=leader_client,
                reason=reason,
            )
            return res
        return add_code_request_to_project(
            project=self,
            code=obj,
            client=client,
            reason=reason,
        )

    def get_messages(self) -> list[ProjectMessage | ProjectThreadMessage]:
        return [
            event
            for event in self.events
            if isinstance(event, (ProjectMessage | ProjectThreadMessage))
        ]

    @property
    def messages(self) -> str:
        message_text = ""
        top_messages = self.get_events(types=ProjectMessage)
        for message in top_messages:
            message_text += (
                f"{str(message.creator_verify_key)[0:8]}: {message.message}\n"
            )
            children = self.get_children(message)
            for child in children:
                message_text += (
                    f"> {str(child.creator_verify_key)[0:8]}: {child.message}\n"
                )
        if message_text == "":
            message_text = "No messages"
        return message_text

    def get_last_seq_no(self) -> int:
        return len(self.events)

    def send_message(self, message: str) -> SyftSuccess | SyftError:
        message_event = ProjectMessage(message=message)
        result = self.add_event(message_event)
        if isinstance(result, SyftSuccess):
            return SyftSuccess(message="Message sent successfully")
        return result

    def reply_message(
        self,
        reply: str,
        message: UID | ProjectMessage | ProjectThreadMessage,
    ) -> SyftSuccess | SyftError:
        if isinstance(message, UID):
            if message not in self.event_ids:
                return SyftError(message=f"Message id: {message} not found")
            message = self.event_id_hashmap[message]

        reply_event: ProjectMessage | ProjectThreadMessage
        if isinstance(message, ProjectMessage):
            reply_event = message.reply(reply)
        elif isinstance(message, ProjectThreadMessage):  # type: ignore[unreachable]
            reply_event = ProjectThreadMessage(
                message=reply, parent_event_id=message.parent_event_id
            )
        else:
            return SyftError(
                message=f"You can only reply to a message: {type(message)}"
                "Kindly re-check the msg"
            )

        result = self.add_event(reply_event)
        if isinstance(result, SyftSuccess):
            return SyftSuccess(message="Reply sent successfully")
        return result

    def create_poll(
        self,
        question: str | None = None,
        choices: list[str] | None = None,
    ) -> SyftSuccess | SyftError:
        if (
            question is None
            or choices is None
            or not isinstance(question, str)
            or not isinstance(choices, list)
        ):
            question, choices = poll_creation_wizard()

        poll_event = ProjectMultipleChoicePoll(question=question, choices=choices)
        result = self.add_event(poll_event)
        if isinstance(result, SyftSuccess):
            return SyftSuccess(message="Poll created successfully")
        return result

    def answer_poll(
        self,
        poll: UID | ProjectMultipleChoicePoll,
        answer: int | None = None,
    ) -> SyftSuccess | SyftError:
        if isinstance(poll, UID):
            if poll not in self.event_ids:
                return SyftError(message=f"Poll id: {poll} not found")
            poll = self.event_id_hashmap[poll]

        if not isinstance(poll, ProjectMultipleChoicePoll):
            return SyftError(  # type: ignore[unreachable]
                message=f"You can only reply to a poll: {type(poll)}"
                "Kindly re-check the poll"
            )

        if not isinstance(answer, int) or answer <= 0 or answer > len(poll.choices):
            answer = poll_answer_wizard(poll)

        answer_event = poll.answer(answer)

        result = self.add_event(answer_event)
        if isinstance(result, SyftSuccess):
            return SyftSuccess(message="Poll answered successfully")
        return result

    def add_request(
        self,
        request: Request,
    ) -> SyftSuccess | SyftError:
        linked_request = LinkedObject.from_obj(request, server_uid=request.server_uid)
        request_event = ProjectRequest(linked_request=linked_request)
        result = self.add_event(request_event)

        if isinstance(result, SyftSuccess):
            return SyftSuccess(message="Request created successfully")
        return result

    # Since currently we do not have the notion of denying a request
    # Adding only approve request, which would later be used to approve or deny a request
    def approve_request(
        self,
        request: UID | ProjectRequest,
    ) -> SyftError | SyftSuccess:
        if isinstance(request, UID):
            if request not in self.event_ids:
                return SyftError(message=f"Request id: {request} not found")
            request = self.event_id_hashmap[request]

        request_event: ProjectRequestResponse
        if isinstance(request, ProjectRequest):
            request_event = request.approve()
            if isinstance(request_event, SyftError):
                return request_event
        else:
            return SyftError(  # type: ignore[unreachable]
                message=f"You can only approve a request: {type(request)}"
                "Kindly re-check the request"
            )
        result = self.add_event(request_event)
        if isinstance(result, SyftSuccess):
            return SyftSuccess(message="Request approved successfully")
        return result

    def sync(self, verbose: bool | None = True) -> SyftSuccess | SyftError:
        """Sync the latest project with the state sync leader"""

        leader_client = self.get_leader_client(self.user_signing_key)

        unsynced_events = leader_client.api.services.project.sync(
            project_id=self.id, seq_no=self.get_last_seq_no()
        )
        if isinstance(unsynced_events, SyftError):
            return unsynced_events

        # UI progress bar for syncing
        if verbose and unsynced_events:
            with Progress() as progress:
                curr_val = 0
                task1 = progress.add_task(
                    f"[bold white]Syncing... {curr_val}/{len(unsynced_events)}",
                    total=len(unsynced_events),
                )

                while not progress.finished:
                    event = unsynced_events[curr_val]
                    curr_val += 1
                    progress.tasks[
                        task1
                    ].description = (
                        f"[bold white]Syncing... {curr_val}/{len(unsynced_events)}"
                    )
                    progress.update(task1, advance=1)
                    self.events.append(event)
                    self.event_id_hashmap[event.id] = event
                    # for a fancy UI view , deliberately slowing the sync
                    if curr_val <= 7:
                        time.sleep(0.1)
        else:
            for event in unsynced_events:
                self.events.append(event)
                self.event_id_hashmap[event.id] = event

        return SyftSuccess(message="Synced project  with Leader")

    @property
    def requests(self) -> list[Request]:
        return [
            event.request for event in self.events if isinstance(event, ProjectRequest)
        ]

    @property
    def pending_requests(self) -> int:
        return sum(
            [request.status == RequestStatus.PENDING for request in self.requests]
        )


@serializable(without=["bootstrap_events", "clients"])
class ProjectSubmit(SyftObject):
    __canonical_name__ = "ProjectSubmit"
    __version__ = SYFT_OBJECT_VERSION_1

    __hash_exclude_attrs__ = [
        "start_hash",
        "users",
        "members",
        "clients",
        "leader_server_route",
        "bootstrap_events",
    ]

    # stash rules
    __repr_attrs__ = ["name", "description", "created_by"]
    __attr_unique__ = ["name"]

    id: UID

    # Init args
    name: str
    description: str | None = None
    members: list[SyftClient] | list[ServerIdentity]

    # These will be automatically populated
    users: list[UserIdentity] = []
    created_by: str | None = None
    username: str | None = None
    clients: list[SyftClient] = []  # List of member clients
    start_hash: str = ""

    # Project sync args
    leader_server_route: ServerRoute | None = None
    state_sync_leader: ServerIdentity | None = None
    bootstrap_events: list[ProjectEvent] | None = []

    # Unused at the moment
    project_permissions: set[str] = set()
    consensus_model: ConsensusModel = DemocraticConsensusModel()

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)

        # Preserve member SyftClients in a private variable clients
        # self.members will be List[ServerIdentity] on the server i.e. self.clients = []
        self.clients = self.get_syft_clients(self.members)

        # If ProjectSubmit is being re-created at server side
        if len(self.clients) == 0:
            return

        # Populate the created_by
        self.created_by = self.clients[0].logged_in_user

        # Extract information of logged in user from syft clients
        self.users = [UserIdentity.from_client(client) for client in self.clients]

        # Assign logged in user name as project creator
        if isinstance(self.clients[0].me, UserView):
            self.username = self.clients[0].me.name
        else:
            self.username = ""

        # Convert SyftClients to ServerIdentities
        self.members = list(map(self.to_server_identity, self.members))

    def _repr_html_(self) -> Any:
        return (
            f"""
            <style>
            .syft-project-create {{color: {SURFACE[options.color_theme]};}}
            </style>
            """
            + "<div class='syft-project-create'>"
            + f"<h3>{self.name}</h3>"
            + f"<p>{self.description}</p>"
            + f"<p><strong>Created by: </strong>{self.username} ({self.created_by})</p>"
            + "</div>"
        )

    @field_validator("members", mode="before")
    @classmethod
    def verify_members(
        cls, val: list[SyftClient] | list[ServerIdentity]
    ) -> list[SyftClient] | list[ServerIdentity]:
        # SyftClients must be logged in by the same emails
        clients = cls.get_syft_clients(val)
        if len(clients) > 0:
            emails = {client.logged_in_user for client in clients}
            if len(emails) > 1:
                raise ValueError(
                    f"All clients must be logged in from the same account. Found multiple: {emails}"
                )
        return val

    @staticmethod
    def get_syft_clients(
        vals: list[SyftClient] | list[ServerIdentity],
    ) -> list[SyftClient]:
        return [client for client in vals if isinstance(client, SyftClient)]

    @staticmethod
    def to_server_identity(val: SyftClient | ServerIdentity) -> ServerIdentity:
        if isinstance(val, ServerIdentity):
            return val
        elif isinstance(val, SyftClient) and val.metadata is not None:
            metadata = val.metadata.to(ServerMetadata)
            return metadata.to(ServerIdentity)
        else:
            raise SyftException(
                f"members must be SyftClient or ServerIdentity. Received: {type(val)}"
            )

    def create_code_request(
        self, obj: SubmitUserCode, client: SyftClient, reason: str | None = None
    ) -> SyftError | SyftSuccess:
        return add_code_request_to_project(
            project=self,
            code=obj,
            client=client,
            reason=reason,
        )

    @deprecated(
        reason="Project.start has been renamed to Project.send", return_syfterror=True
    )
    def start(self, return_all_projects: bool = False) -> Project | list[Project]:
        return self.send(return_all_projects=return_all_projects)

    def send(self, return_all_projects: bool = False) -> Project | list[Project]:
        # Currently we are assuming that the first member is the leader
        # This would be changed in our future leaderless approach
        leader = self.clients[0]
        followers = self.clients[1:]

        try:
            # Check for DS role across all members
            self._pre_submit_checks(self.clients)

            # Exchange route between leaders and followers
            self._exchange_routes(leader, followers)

            # create project for each server
            projects_map = self._create_projects(self.clients)

            # bootstrap project with pending events on leader server's project
            self._bootstrap_events(projects_map[leader])

            if return_all_projects:
                return list(projects_map.values())

            return projects_map[leader]
        except SyftException as exp:
            return SyftError(message=str(exp))

    def _pre_submit_checks(self, clients: list[SyftClient]) -> bool:
        try:
            # Check if the user can create projects
            for client in clients:
                result = client.api.services.project.can_create_project()
                if isinstance(result, SyftError):
                    raise SyftException(result.message)
        except Exception:
            raise SyftException("Only Data Scientists can create projects")

        return True

    def _exchange_routes(self, leader: SyftClient, followers: list[SyftClient]) -> None:
        # Since we are implementing a leader based system
        # To be able to optimize exchanging routes.
        # We require only the leader to exchange routes with all the members
        # Meaning if we could guarantee, that the leader server is able to reach the members
        # the project events could be broadcasted to all the members

        for follower in followers:
            result = leader.exchange_route(follower)
            if isinstance(result, SyftError):
                raise SyftException(result.message)

        self.leader_server_route = connection_to_route(leader.connection)

    def _create_projects(self, clients: list[SyftClient]) -> dict[SyftClient, Project]:
        projects: dict[SyftClient, Project] = {}

        for client in clients:
            result = client.api.services.project.create_project(project=self)
            if isinstance(result, SyftError):
                raise SyftException(result.message)
            projects[client] = result

        return projects

    def _bootstrap_events(self, leader_project: Project) -> None:
        if not self.bootstrap_events:
            return

        while len(self.bootstrap_events) > 0:
            event = self.bootstrap_events.pop(0)
            result = leader_project.add_event(event)
            if isinstance(result, SyftError):
                raise SyftException(result.message)


def add_members_as_owners(members: list[SyftVerifyKey]) -> set[str]:
    keys = set()
    for member in members:
        owner_key = f"OWNER_{member.verify_key}"
        keys.add(owner_key)
    return keys


def elect_leader(context: TransformContext) -> TransformContext:
    if context.output is not None:
        if len(context.output["members"]) == 0:
            raise ValueError("Project's require at least one member")
        context.output["state_sync_leader"] = context.output["members"][0]

    return context


def check_permissions(context: TransformContext) -> TransformContext:
    if context.output is None:
        return context

    if len(context.output["members"]) > 1:
        # more than 1 server
        pass
    # check at least one owner
    if len(context.output["project_permissions"]) == 0:
        project_permissions = context.output["project_permissions"]
        project_permissions = project_permissions.union(
            add_members_as_owners(context.output["members"])
        )
        context.output["project_permissions"] = project_permissions

    return context


def add_creator_name(context: TransformContext) -> TransformContext:
    if context.output is not None and context.obj is not None:
        context.output["username"] = context.obj.username
    return context


@transform(ProjectSubmit, Project)
def new_projectsubmit_to_project() -> list[Callable]:
    return [elect_leader, check_permissions, add_creator_name]


def hash_object(obj: Any) -> tuple[bytes, str]:
    """Hashes an object using sha256

    Args:
        obj (Any): Object to be hashed

    Returns:
        str: Hashed value of the object
    """
    hash_bytes = _serialize(obj, to_bytes=True, for_hashing=True)
    hash = hashlib.sha256(hash_bytes)
    return (hash.digest(), hash.hexdigest())


def create_project_hash(project: Project) -> tuple[bytes, str]:
    # Creating a custom hash for the project
    # as the recursive hash is yet to be revamped
    # for primitives python types

    # hashing is calculated based on the following attributes
    # attrs = ["name", "description", "created_by", "members", "users"]

    return hash_object(
        [
            project.name,
            project.description,
            project.created_by,
            [hash_object(member) for member in project.members],
            [hash_object(user) for user in project.users],
        ]
    )


def create_project_event_hash(project_event: ProjectEvent) -> tuple[bytes, str]:
    # Creating a custom hash for the project
    # as the recursive hash is yet to be revamped
    # for primitives python types.

    # hashing is calculated based on the following attributes,
    # attrs = ["id", "project_id", "seq no",
    # "prev_event_uid", "prev_event_hash", "creator_verify_key"]

    return hash_object(
        [
            project_event.id,
            project_event.project_id,
            project_event.seq_no,
            project_event.prev_event_uid,
            project_event.timestamp.utc_timestamp,
            project_event.prev_event_hash,
            hash_object(project_event.creator_verify_key)[1],
        ]
    )



# stdlib
from collections.abc import Callable
import logging
import os
from typing import Any
from typing import TypeVar

logger = logging.getLogger(__name__)


def str_to_bool(bool_str: str | None) -> bool:
    result = False
    bool_str = str(bool_str).lower()
    if bool_str == "true" or bool_str == "1":
        result = True
    return result


TRACE_MODE = str_to_bool(os.environ.get("TRACE", "False"))


T = TypeVar("T", bound=Callable | type)


def noop(__func_or_class: T, /, *args: Any, **kwargs: Any) -> T:
    return __func_or_class


if not TRACE_MODE:
    instrument = noop
else:
    try:
        service_name = os.environ.get("SERVICE_NAME", "client")
        jaeger_host = os.environ.get("JAEGER_HOST", "localhost")
        jaeger_port = int(os.environ.get("JAEGER_PORT", "14268"))

        # third party
        from opentelemetry import trace
        from opentelemetry.exporter.jaeger.thrift import JaegerExporter
        from opentelemetry.sdk.resources import Resource
        from opentelemetry.sdk.resources import SERVICE_NAME
        from opentelemetry.sdk.trace import TracerProvider
        from opentelemetry.sdk.trace.export import BatchSpanProcessor

        trace.set_tracer_provider(
            TracerProvider(resource=Resource.create({SERVICE_NAME: service_name}))
        )
        jaeger_exporter = JaegerExporter(
            # agent_host_name=jaeger_host,
            # agent_port=jaeger_port,
            collector_endpoint=f"http://{jaeger_host}:{jaeger_port}/api/traces?format=jaeger.thrift",
            # udp_split_oversized_batches=True,
        )

        trace.get_tracer_provider().add_span_processor(
            BatchSpanProcessor(jaeger_exporter)
        )

        # from opentelemetry.sdk.trace.export import ConsoleSpanExporter
        # console_exporter = ConsoleSpanExporter()
        # span_processor = BatchSpanProcessor(console_exporter)
        # trace.get_tracer_provider().add_span_processor(span_processor)

        # third party
        import opentelemetry.instrumentation.requests

        opentelemetry.instrumentation.requests.RequestsInstrumentor().instrument()

        # relative
        # from opentelemetry.instrumentation.digma.trace_decorator import (
        #     instrument as _instrument,
        # )
        #
        # until this is merged:
        # https://github.com/digma-ai/opentelemetry-instrumentation-digma/pull/41
        from .trace_decorator import instrument as _instrument

        instrument = _instrument
    except Exception as e:
        logger.error("Failed to import opentelemetry", exc_info=e)
        instrument = noop



# stdlib
from typing import Any

# relative
from ..serde.serializable import serializable
from .base import SyftBaseModel


@serializable(canonical_name="CachedSyftObject", version=1)
class CachedSyftObject(SyftBaseModel):
    """This class is used to represent the cached result."""

    result: Any
    error_msg: str | None = None



# stdlib
import contextlib

# third party
import docker
import pydantic

# relative
from ...custom_worker.config import PrebuiltWorkerConfig
from ...custom_worker.config import WorkerConfig
from ...custom_worker.k8s import IN_KUBERNETES
from ...serde.serializable import serializable
from ...store.document_store import DocumentStore
from ...types.datetime import DateTime
from ...types.dicttuple import DictTuple
from ...types.uid import UID
from ..context import AuthedServiceContext
from ..response import SyftError
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import service_method
from ..user.user_roles import DATA_OWNER_ROLE_LEVEL
from ..user.user_roles import DATA_SCIENTIST_ROLE_LEVEL
from .image_registry import SyftImageRegistry
from .image_registry_service import SyftImageRegistryService
from .utils import image_build
from .utils import image_push
from .worker_image import SyftWorkerImage
from .worker_image import SyftWorkerImageIdentifier
from .worker_image_stash import SyftWorkerImageStash


@serializable(canonical_name="SyftWorkerImageService", version=1)
class SyftWorkerImageService(AbstractService):
    store: DocumentStore
    stash: SyftWorkerImageStash

    def __init__(self, store: DocumentStore) -> None:
        self.store = store
        self.stash = SyftWorkerImageStash(store=store)

    @service_method(
        path="worker_image.submit",
        name="submit",
        roles=DATA_OWNER_ROLE_LEVEL,
    )
    def submit(
        self, context: AuthedServiceContext, worker_config: WorkerConfig
    ) -> SyftSuccess | SyftError:
        image_identifier: SyftWorkerImageIdentifier | None = None
        if isinstance(worker_config, PrebuiltWorkerConfig):
            try:
                image_identifier = SyftWorkerImageIdentifier.from_str(worker_config.tag)
            except Exception:
                return SyftError(
                    f"Invalid Docker image name: {worker_config.tag}.\n"
                    + "Please specify the image name in this format <registry>/<repo>:<tag>."
                )

        worker_image = SyftWorkerImage(
            config=worker_config,
            created_by=context.credentials,
            image_identifier=image_identifier,
        )
        res = self.stash.set(context.credentials, worker_image)

        if res.is_err():
            return SyftError(message=res.err())

        return SyftSuccess(
            message=f"Dockerfile ID: {worker_image.id} successfully submitted."
        )

    @service_method(
        path="worker_image.build",
        name="build",
        roles=DATA_OWNER_ROLE_LEVEL,
    )
    def build(
        self,
        context: AuthedServiceContext,
        image_uid: UID,
        tag: str,
        registry_uid: UID | None = None,
        pull_image: bool = True,
    ) -> SyftSuccess | SyftError:
        registry: SyftImageRegistry | None = None

        if IN_KUBERNETES and registry_uid is None:
            return SyftError(message="Registry UID is required in Kubernetes mode.")

        result = self.stash.get_by_uid(credentials=context.credentials, uid=image_uid)
        if result.is_err():
            return SyftError(
                message=f"Failed to get image for uid: {image_uid}. Error: {result.err()}"
            )

        worker_image: SyftWorkerImage = result.ok()

        if registry_uid:
            # get registry from image registry service
            image_registry_service: AbstractService = context.server.get_service(
                SyftImageRegistryService
            )
            registry_result = image_registry_service.get_by_id(context, registry_uid)
            if registry_result.is_err():
                return registry_result
            registry = registry_result.ok()

        try:
            if registry:
                image_identifier = SyftWorkerImageIdentifier.with_registry(
                    tag=tag, registry=registry
                )
            else:
                image_identifier = SyftWorkerImageIdentifier.from_str(tag=tag)
        except pydantic.ValidationError as e:
            return SyftError(message=f"Failed to create tag: {e}")

        # if image is already built and identifier is unchanged, return an error
        if (
            worker_image.built_at
            and worker_image.image_identifier
            and worker_image.image_identifier.full_name_with_tag
            == image_identifier.full_name_with_tag
        ):
            return SyftError(message=f"Image ID: {image_uid} is already built")

        worker_image.image_identifier = image_identifier
        result = None

        if not context.server.in_memory_workers:
            build_result = image_build(worker_image, pull=pull_image)
            if isinstance(build_result, SyftError):
                return build_result

            worker_image.image_hash = build_result.image_hash
            worker_image.built_at = DateTime.now()

            result = SyftSuccess(
                message=f"Build for Worker ID: {worker_image.id} succeeded.\n{build_result.logs}"
            )
        else:
            result = SyftSuccess(
                message="Image building skipped, since using in-memory workers."
            )

        update_result = self.stash.update(context.credentials, obj=worker_image)

        if update_result.is_err():
            return SyftError(
                message=f"Failed to update image meta information: {update_result.err()}"
            )

        return result

    @service_method(
        path="worker_image.push",
        name="push",
        roles=DATA_OWNER_ROLE_LEVEL,
    )
    def push(
        self,
        context: AuthedServiceContext,
        image_uid: UID,
        username: str | None = None,
        password: str | None = None,
    ) -> SyftSuccess | SyftError:
        result = self.stash.get_by_uid(credentials=context.credentials, uid=image_uid)
        if result.is_err():
            return SyftError(
                message=f"Failed to get Image ID: {image_uid}. Error: {result.err()}"
            )
        worker_image: SyftWorkerImage = result.ok()

        if not worker_image.is_built:
            return SyftError(message=f"Image ID: {worker_image.id} is not built yet.")
        elif (
            worker_image.image_identifier is None
            or worker_image.image_identifier.registry_host == ""
        ):
            return SyftError(
                message=f"Image ID: {worker_image.id} does not have a valid registry host."
            )

        result = image_push(
            image=worker_image,
            username=username,
            password=password,
        )

        if isinstance(result, SyftError):
            return result

        return SyftSuccess(
            message=f'Pushed Image ID: {worker_image.id} to "{worker_image.image_identifier.full_name_with_tag}".'
        )

    @service_method(
        path="worker_image.get_all",
        name="get_all",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def get_all(
        self, context: AuthedServiceContext
    ) -> DictTuple[str, SyftWorkerImage] | SyftError:
        """
        One image one docker file for now
        """
        result = self.stash.get_all(credentials=context.credentials)
        if result.is_err():
            return SyftError(message=f"{result.err()}")
        images: list[SyftWorkerImage] = result.ok()

        res = {}
        # if image is built, index it by full_name_with_tag
        for im in images:
            if im.is_built and im.image_identifier is not None:
                res[im.image_identifier.full_name_with_tag] = im
        # and then index all images by id
        # TODO: jupyter repr needs to be updated to show unique values
        # (even if multiple keys point to same value)
        res.update({im.id.to_string(): im for im in images if not im.is_built})

        return DictTuple(res)

    @service_method(
        path="worker_image.remove",
        name="remove",
        roles=DATA_OWNER_ROLE_LEVEL,
    )
    def remove(
        self, context: AuthedServiceContext, uid: UID
    ) -> SyftSuccess | SyftError:
        #  Delete Docker image given image tag
        res = self.stash.get_by_uid(credentials=context.credentials, uid=uid)
        if res.is_err():
            return SyftError(message=f"{res.err()}")
        image: SyftWorkerImage = res.ok()

        if context.server.in_memory_workers:
            pass
        elif IN_KUBERNETES:
            # TODO: Implement image deletion in kubernetes
            return SyftError(
                message="Image Deletion is not yet implemented in Kubernetes !!"
            )
        elif image and image.image_identifier:
            try:
                full_tag: str = image.image_identifier.full_name_with_tag
                with contextlib.closing(docker.from_env()) as client:
                    client.images.remove(image=full_tag)
            except docker.errors.ImageNotFound:
                return SyftError(message=f"Image Tag: {full_tag} not found.")
            except Exception as e:
                return SyftError(
                    message=f"Failed to delete Image Tag: {full_tag}. Error: {e}"
                )

        result = self.stash.delete_by_uid(credentials=context.credentials, uid=uid)

        if result.is_err():
            return SyftError(message=f"{result.err()}")

        returned_message: str = (
            result.ok().message + f". Image ID: {uid} deleted successfully."
        )

        return SyftSuccess(message=returned_message)

    @service_method(
        path="worker_image.get_by_uid",
        name="get_by_uid",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def get_by_uid(
        self, context: AuthedServiceContext, uid: UID
    ) -> SyftWorkerImage | SyftError:
        res = self.stash.get_by_uid(credentials=context.credentials, uid=uid)
        if res.is_err():
            return SyftError(
                message=f"Failed to get image with uid {uid}. Error: {res.err()}"
            )
        image: SyftWorkerImage = res.ok()
        return image

    @service_method(
        path="worker_image.get_by_config",
        name="get_by_config",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def get_by_config(
        self, context: AuthedServiceContext, worker_config: WorkerConfig
    ) -> SyftWorkerImage | SyftError:
        res = self.stash.get_by_worker_config(
            credentials=context.credentials, config=worker_config
        )
        if res.is_err():
            return SyftError(
                message=f"Failed to get image with docker config {worker_config}. Error: {res.err()}"
            )
        image: SyftWorkerImage = res.ok()
        return image



# stdlib
from collections.abc import Callable
from enum import Enum
from enum import EnumMeta
import os
import tempfile
import types
from typing import Any

# third party
from capnp.lib.capnp import _DynamicStructBuilder
from pydantic import BaseModel

# syft absolute
import syft as sy

# relative
from ..types.syft_object_registry import SyftObjectRegistry
from .capnp import get_capnp_schema
from .util import compatible_with_large_file_writes_capnp

TYPE_BANK = {}  # type: ignore
SYFT_CLASSES_MISSING_CANONICAL_NAME = []

recursive_scheme = get_capnp_schema("recursive_serde.capnp").RecursiveSerde

SPOOLED_FILE_MAX_SIZE_SERDE = 50 * (1024**2)  # 50MB


def get_types(cls: type, keys: list[str] | None = None) -> list[type] | None:
    if keys is None:
        return None
    types = []
    for key in keys:
        _type = None
        annotations = getattr(cls, "__annotations__", None)
        if annotations and key in annotations:
            _type = annotations[key]
        else:
            for parent_cls in cls.mro():
                sub_annotations = getattr(parent_cls, "__annotations__", None)
                if sub_annotations and key in sub_annotations:
                    _type = sub_annotations[key]
        if _type is None:
            return None
        types.append(_type)
    return types


def check_fqn_alias(cls: object | type) -> tuple[str, ...] | None:
    """Currently, typing.Any has different metaclasses in different versions of Python 🤦‍♂️.
    For Python <=3.10
    Any is an instance of typing._SpecialForm

    For Python >=3.11
    Any is an instance of typing._AnyMeta
    Hence adding both the aliases to the type bank.

    This would cause issues, when the server and client
    have different python versions.

    As their serde is same, we can use the same serde for both of them.
    with aliases for  fully qualified names in type bank

    In a similar manner for Enum.

    For Python<=3.10:
    Enum is metaclass of enum.EnumMeta

    For Python>=3.11:
    Enum is metaclass of enum.EnumType
    """
    if cls == Any:
        return ("typing._AnyMeta", "typing._SpecialForm")
    if cls == EnumMeta:
        return ("enum.EnumMeta", "enum.EnumType")

    return None


def has_canonical_name_version(
    cls: type, cannonical_name: str | None, version: int | None
) -> bool:
    cls_canonical_name = getattr(cls, "__canonical_name__", None)
    cls_version = getattr(cls, "__version__", None)
    return bool(cls_canonical_name or cannonical_name) and bool(cls_version or version)


def validate_cannonical_name_version(
    cls: type, canonical_name: str | None, version: int | None
) -> tuple[str, int]:
    cls_canonical_name = getattr(cls, "__canonical_name__", None)
    cls_version = getattr(cls, "__version__", None)
    if cls_canonical_name and canonical_name:
        raise ValueError(
            "Cannot specify both __canonical_name__ attribute and cannonical_name argument."
        )
    if cls_version and version:
        raise ValueError(
            "Cannot specify both __version__ attribute and version argument."
        )
    if cls_canonical_name is None and canonical_name is None:
        raise ValueError(
            "Must specify either __canonical_name__ attribute or cannonical_name argument."
        )
    if cls_version is None and version is None:
        raise ValueError(
            "Must specify either __version__ attribute or version argument."
        )

    canonical_name = canonical_name or cls_canonical_name
    version = version or cls_version
    return canonical_name, version  # type: ignore


def skip_unregistered_class(
    cls: type, canonical_name: str | None, version: str | None
) -> bool:
    """
    Used to gather all classes that are missing canonical_name and version for development.

    Returns True if the class should be skipped, False otherwise.
    """

    search_unregistered_classes = (
        os.getenv("SYFT_SEARCH_MISSING_CANONICAL_NAME", False) == "true"
    )
    if not search_unregistered_classes:
        return False
    if not has_canonical_name_version(cls, canonical_name, version):
        if cls.__module__.startswith("syft."):
            SYFT_CLASSES_MISSING_CANONICAL_NAME.append(cls)
            return True
    return False


def recursive_serde_register(
    cls: object | type,
    serialize: Callable | None = None,
    deserialize: Callable | None = None,
    serialize_attrs: list | None = None,
    exclude_attrs: list | None = None,
    inherit_attrs: bool | None = True,
    inheritable_attrs: bool | None = True,
    canonical_name: str | None = None,
    version: int | None = None,
) -> None:
    pydantic_fields = None
    base_attrs = None
    attribute_list: set[str] = set()

    cls = type(cls) if not isinstance(cls, type) else cls

    if skip_unregistered_class(cls, canonical_name, version):
        return

    canonical_name, version = validate_cannonical_name_version(
        cls, canonical_name, version
    )

    nonrecursive = bool(serialize and deserialize)
    _serialize = serialize if nonrecursive else rs_object2proto
    _deserialize = deserialize if nonrecursive else rs_proto2object
    is_pydantic = issubclass(cls, BaseModel)
    hash_exclude_attrs = getattr(cls, "__hash_exclude_attrs__", [])

    if inherit_attrs and not is_pydantic:
        # get attrs from base class
        base_attrs = getattr(cls, "__syft_serializable__", [])
        attribute_list.update(base_attrs)

    if is_pydantic and not serialize_attrs:
        # if pydantic object and attrs are provided, the get attrs from __fields__
        # cls.__fields__ auto inherits attrs
        pydantic_fields = [
            field
            for field, field_info in cls.model_fields.items()
            if not (
                field_info.annotation is not None
                and hasattr(field_info.annotation, "__origin__")
                and field_info.annotation.__origin__
                in (Callable, types.FunctionType, types.LambdaType)
            )
        ]
        attribute_list.update(pydantic_fields)

    if serialize_attrs:
        # If serialize_attrs is provided, append it to our attr list
        attribute_list.update(serialize_attrs)

    if issubclass(cls, Enum):
        attribute_list.update(["value"])

    exclude_attrs = [] if exclude_attrs is None else exclude_attrs
    attribute_list = (
        attribute_list - set(exclude_attrs) - {"syft_pre_hooks__", "syft_post_hooks__"}
    )

    if inheritable_attrs and attribute_list and not is_pydantic:
        # only set __syft_serializable__ for non-pydantic classes because
        # pydantic objects inherit by default
        cls.__syft_serializable__ = attribute_list

    attributes = set(attribute_list) if attribute_list else None
    attribute_types = get_types(cls, attributes)
    serde_overrides = getattr(cls, "__serde_overrides__", {})

    # without fqn duplicate class names overwrite
    serde_attributes = (
        nonrecursive,
        _serialize,
        _deserialize,
        attributes,
        exclude_attrs,
        serde_overrides,
        hash_exclude_attrs,
        cls,
        attribute_types,
        version,
    )

    SyftObjectRegistry.register_cls(canonical_name, version, serde_attributes)

    alias_fqn = check_fqn_alias(cls)
    if isinstance(alias_fqn, tuple):
        for alias in alias_fqn:
            alias_canonical_name = canonical_name + f"_{alias}"
            SyftObjectRegistry.register_cls(alias_canonical_name, 1, serde_attributes)


def chunk_bytes(
    field_obj: Any,
    ser_func: Callable,
    field_name: str | int,
    builder: _DynamicStructBuilder,
) -> None:
    data = ser_func(field_obj)
    size_of_data = len(data)
    if compatible_with_large_file_writes_capnp(size_of_data):
        with tempfile.TemporaryFile() as tmp_file:
            # Write data to a file to save RAM
            tmp_file.write(data)
            tmp_file.seek(0)
            del data

            CHUNK_SIZE = int(5.12e8)  # capnp max for a List(Data) field
            list_size = size_of_data // CHUNK_SIZE + 1
            data_lst = builder.init(field_name, list_size)
            for idx in range(list_size):
                bytes_to_read = min(CHUNK_SIZE, size_of_data)
                data_lst[idx] = tmp_file.read(bytes_to_read)
                size_of_data -= CHUNK_SIZE
    else:
        CHUNK_SIZE = int(5.12e8)  # capnp max for a List(Data) field
        list_size = len(data) // CHUNK_SIZE + 1
        data_lst = builder.init(field_name, list_size)
        END_INDEX = CHUNK_SIZE
        for idx in range(list_size):
            START_INDEX = idx * CHUNK_SIZE
            END_INDEX = min(START_INDEX + CHUNK_SIZE, len(data))
            data_lst[idx] = data[START_INDEX:END_INDEX]


def combine_bytes(capnp_list: list[bytes]) -> bytes:
    # TODO: make sure this doesn't copy, perhaps allocate a fixed size buffer
    # and move the bytes into it as we go
    bytes_value = b""
    for value in capnp_list:
        bytes_value += value
    return bytes_value


def rs_object2proto(self: Any, for_hashing: bool = False) -> _DynamicStructBuilder:
    # relative
    from ..types.syft_object import DYNAMIC_SYFT_ATTRIBUTES

    is_type = False
    if isinstance(self, type):
        is_type = True

    msg = recursive_scheme.new_message()

    # todo: rewrite and make sure every object has a canonical name and version
    canonical_name, version = SyftObjectRegistry.get_canonical_name_version(self)

    if not SyftObjectRegistry.has_serde_class(canonical_name, version):
        # third party
        raise Exception(
            f"obj2proto: {canonical_name} version {version} not in SyftObjectRegistry"
        )

    msg.canonicalName = canonical_name
    msg.version = version

    (
        nonrecursive,
        serialize,
        _,
        attribute_list,
        exclude_attrs_list,
        serde_overrides,
        hash_exclude_attrs,
        _,
        _,
        _,
    ) = SyftObjectRegistry.get_serde_properties(canonical_name, version)

    if nonrecursive or is_type:
        if serialize is None:
            raise Exception(
                f"Cant serialize {type(self)} nonrecursive without serialize."
            )
        chunk_bytes(self, serialize, "nonrecursiveBlob", msg)
        return msg

    if attribute_list is None:
        attribute_list = self.__dict__.keys()

    hash_exclude_attrs_set = (
        set(hash_exclude_attrs).union(set(DYNAMIC_SYFT_ATTRIBUTES))
        if for_hashing
        else set()
    )
    attribute_list = (
        set(attribute_list) - set(exclude_attrs_list) - hash_exclude_attrs_set
    )

    msg.init("fieldsName", len(attribute_list))
    msg.init("fieldsData", len(attribute_list))

    for idx, attr_name in enumerate(sorted(attribute_list)):
        if not hasattr(self, attr_name):
            raise ValueError(
                f"{attr_name} on {type(self)} does not exist, serialization aborted!"
            )

        field_obj = getattr(self, attr_name)
        transforms = serde_overrides.get(attr_name, None)

        if transforms is not None:
            field_obj = transforms[0](field_obj)

        if isinstance(field_obj, types.FunctionType):
            continue

        msg.fieldsName[idx] = attr_name
        chunk_bytes(
            field_obj,
            lambda x: sy.serialize(x, to_bytes=True, for_hashing=for_hashing),
            idx,
            msg.fieldsData,
        )

    return msg


def rs_bytes2object(blob: bytes) -> Any:
    MAX_TRAVERSAL_LIMIT = 2**64 - 1

    with recursive_scheme.from_bytes(
        blob, traversal_limit_in_words=MAX_TRAVERSAL_LIMIT
    ) as msg:
        return rs_proto2object(msg)


def map_fqns_for_backward_compatibility(fqn: str) -> str:
    """for backwards compatibility with 0.8.6. Sometimes classes where moved to another file. Which is
    exactly why we are implementing it differently"""
    mapping = {
        "syft.service.dataset.dataset.MarkdownDescription": "syft.util.misc_objs.MarkdownDescription",
        "syft.service.object_search.object_migration_state.SyftObjectMigrationState": "syft.service.migration.object_migration_state.SyftObjectMigrationState",  # noqa: E501
    }
    if fqn in mapping:
        return mapping[fqn]
    else:
        return fqn


def rs_proto2object(proto: _DynamicStructBuilder) -> Any:
    # relative
    from .deserialize import _deserialize

    class_type: type | Any = type(None)

    canonical_name = proto.canonicalName
    version = getattr(proto, "version", -1)

    if not SyftObjectRegistry.has_serde_class(canonical_name, version):
        # third party
        raise Exception(
            f"proto2obj: {canonical_name} version {version} not in SyftObjectRegistry"
        )

    # TODO: 🐉 sort this out, basically sometimes the syft.user classes are not in the
    # module name space in sub-processes or threads even though they are loaded on start
    # its possible that the uvicorn awsgi server is preloading a bunch of threads
    # however simply getting the class from the TYPE_BANK doesn't always work and
    # causes some errors so it seems like we want to get the local one where possible
    (
        nonrecursive,
        _,
        deserialize,
        _,
        _,
        serde_overrides,
        _,
        cls,
        _,
        version,
    ) = SyftObjectRegistry.get_serde_properties(canonical_name, version)

    class_type = cls

    if nonrecursive:
        if deserialize is None:
            raise Exception(
                f"Cant serialize {type(proto)} nonrecursive without serialize."
            )

        return deserialize(combine_bytes(proto.nonrecursiveBlob))

    kwargs = {}

    for attr_name, attr_bytes_list in zip(proto.fieldsName, proto.fieldsData):
        if attr_name != "":
            attr_bytes = combine_bytes(attr_bytes_list)
            attr_value = _deserialize(attr_bytes, from_bytes=True)
            transforms = serde_overrides.get(attr_name, None)

            if transforms is not None:
                attr_value = transforms[1](attr_value)
            kwargs[attr_name] = attr_value

    if hasattr(class_type, "serde_constructor"):
        return class_type.serde_constructor(kwargs)

    if issubclass(class_type, Enum) and "value" in kwargs:
        obj = class_type.__new__(class_type, kwargs["value"])
    elif issubclass(class_type, BaseModel):
        # if we skip the __new__ flow of BaseModel we get the error
        # AttributeError: object has no attribute '__fields_set__'

        # if "syft.user" in proto.fullyQualifiedName:
        #     # weird issues with pydantic and ForwardRef on user classes being inited
        #     # with custom state args / kwargs
        #     obj = class_type()
        #     for attr_name, attr_value in kwargs.items():
        #         setattr(obj, attr_name, attr_value)
        # else:
        #     obj = class_type(**kwargs)
        obj = class_type(**kwargs)

    else:
        obj = class_type.__new__(class_type)  # type: ignore
        for attr_name, attr_value in kwargs.items():
            setattr(obj, attr_name, attr_value)

    return obj


# how else do you import a relative file to execute it?
NOTHING = None



# stdlib
import ast
from collections.abc import Callable
import inspect
from inspect import Signature
import keyword
import linecache
import re
import textwrap
from typing import Any
from typing import cast

# third party
from pydantic import ValidationError
from pydantic import field_validator
from pydantic import model_validator
from result import Err
from result import Ok
from result import Result

# relative
from ...abstract_server import AbstractServer
from ...client.client import SyftClient
from ...serde.serializable import serializable
from ...serde.signature import signature_remove_context
from ...types.syft_object import PartialSyftObject
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ...types.syft_object import SyftObject
from ...types.syncable_object import SyncableSyftObject
from ...types.transforms import TransformContext
from ...types.transforms import generate_action_object_id
from ...types.transforms import generate_id
from ...types.transforms import keep
from ...types.transforms import transform
from ...types.uid import UID
from ...util.misc_objs import MarkdownDescription
from ..context import AuthedServiceContext
from ..response import SyftError
from ..user.user import UserView
from ..user.user_service import UserService

NOT_ACCESSIBLE_STRING = "N / A"


class HelperFunctionSet:
    def __init__(self, helper_functions: dict[str, Callable]) -> None:
        self.helper_functions = helper_functions
        for name, func in helper_functions.items():
            setattr(self, name, func)


class TwinAPIAuthedContext(AuthedServiceContext):
    __canonical_name__ = "TwinAPIAuthedContext"
    __version__ = SYFT_OBJECT_VERSION_1

    user: UserView | None = None
    settings: dict[str, Any] | None = None
    code: HelperFunctionSet | None = None
    state: dict[Any, Any] | None = None
    admin_client: SyftClient | None = None
    user_client: SyftClient | None = None


@serializable()
class TwinAPIContextView(SyftObject):
    __canonical_name__ = "TwinAPIContextView"
    __version__ = SYFT_OBJECT_VERSION_1

    __repr_attrs__ = ["settings", "state", "user"]
    user: UserView
    settings: dict[str, Any]
    state: dict[Any, Any]


def get_signature(func: Callable) -> Signature:
    sig = inspect.signature(func)
    sig = signature_remove_context(sig)
    return sig


def register_fn_in_linecache(fname: str, src: str) -> None:
    """adds a function to linecache, such that inspect.getsource works for functions nested in this function.
    This only works if the same function is compiled under the same filename"""
    lines = [
        line + "\n" for line in src.splitlines()
    ]  # use same splitting method same as linecache 112 (py3.12)
    linecache.cache[fname] = (137, None, lines, fname)


@serializable()
class TwinAPIEndpointView(SyftObject):
    # version
    __canonical_name__ = "CustomAPIView"
    __version__ = SYFT_OBJECT_VERSION_1

    path: str
    action_object_id: UID
    signature: Signature
    access: str = "Public"
    mock_function: str | None = None
    private_function: str | None = None
    description: MarkdownDescription | None = None
    mock_helper_functions: list[str] | None = None
    private_helper_functions: list[str] | None = None
    worker_pool: str | None = None
    endpoint_timeout: int = 60

    __repr_attrs__ = [
        "path",
        "signature",
        "worker_pool",
        "endpoint_timeout",
    ]

    def _coll_repr_(self) -> dict[str, Any]:
        if self.mock_function:
            mock_parsed_code = ast.parse(self.mock_function)
            mock_function_name = [
                server.name
                for server in ast.walk(mock_parsed_code)
                if isinstance(server, ast.FunctionDef)
            ][0]
        else:
            mock_function_name = NOT_ACCESSIBLE_STRING

        if self.private_function:
            private_parsed_code = ast.parse(self.private_function)
            private_function_name = [
                server.name
                for server in ast.walk(private_parsed_code)
                if isinstance(server, ast.FunctionDef)
            ][0]
        else:
            private_function_name = NOT_ACCESSIBLE_STRING

        worker_pool = "UNSET (DEFAULT)"
        if self.worker_pool is not None:
            worker_pool = self.worker_pool
        return {
            "API path": self.path,
            "Signature": self.path + str(self.signature),
            "Access": self.access,
            "Mock Function": mock_function_name,
            "Private Function": private_function_name,
            "Worker Pool": worker_pool,
        }


@serializable()
class Endpoint(SyftObject):
    """Base class to perform basic Endpoint validation for both public/private endpoints."""

    # version
    __canonical_name__ = "CustomApiEndpoint"
    __version__ = SYFT_OBJECT_VERSION_1

    api_code: str
    func_name: str
    settings: dict[str, Any] | None = None
    view_access: bool = True
    helper_functions: dict[str, str] | None = None
    state: dict[Any, Any] | None = None
    signature: Signature

    __exclude_sync_diff_attrs__ = ["state"]

    def __repr__(self) -> str:
        type_name = type(self).__name__
        repr_str = f"""<{type_name}: {self.func_name}>

        {self.api_code}
        """
        return textwrap.dedent(repr_str)

    @field_validator("api_code", check_fields=False)
    @classmethod
    def validate_api_code(cls, api_code: str) -> str:
        valid_code = True
        try:
            ast.parse(api_code)
        except SyntaxError:
            # If the code isn't valid Python syntax
            valid_code = False

        if not valid_code:
            raise ValueError("Code must be a valid Python function.")

        return api_code

    @field_validator("func_name", check_fields=False)
    @classmethod
    def validate_func_name(cls, func_name: str) -> str:
        if not str.isidentifier(func_name) or keyword.iskeyword(func_name):
            raise ValueError("Invalid function name.")
        return func_name

    @field_validator("settings", check_fields=False)
    @classmethod
    def validate_settings(
        cls, settings: dict[str, Any] | None
    ) -> dict[str, Any] | None:
        return settings

    def update_state(self, state: dict[Any, Any]) -> None:
        self.state = state

    def build_internal_context(
        self,
        context: AuthedServiceContext,
        admin_client: SyftClient | None = None,
        user_client: SyftClient | None = None,
    ) -> TwinAPIAuthedContext:
        helper_function_dict: dict[str, Callable] = {}
        self.helper_functions = self.helper_functions or {}
        for helper_name, helper_code in self.helper_functions.items():
            # Create a dictionary to serve as local scope
            local_scope: dict[str, Callable] = {}

            # Execute the function string within the local scope
            exec(helper_code, local_scope)  # nosec
            helper_function_dict[helper_name] = local_scope[helper_name]

        helper_function_set = HelperFunctionSet(helper_function_dict)

        user_service = context.server.get_service("userservice")
        user = user_service.get_current_user(context)

        return TwinAPIAuthedContext(
            credentials=context.credentials,
            role=context.role,
            job_id=context.job_id,
            extra_kwargs=context.extra_kwargs,
            has_execute_permissions=context.has_execute_permissions,
            server=context.server,
            id=context.id,
            settings=self.settings or {},
            code=helper_function_set,
            state=self.state or {},
            user=user,
            admin_client=admin_client,
            user_client=user_client,
        )

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        # relative
        from ... import SyftSigningKey
        from ..context import AuthedServiceContext

        mock_context = AuthedServiceContext(
            server=AbstractServer(), credentials=SyftSigningKey.generate().verify_key
        )
        return self.call_locally(mock_context, *args, **kwargs)

    def call_locally(
        self, context: AuthedServiceContext, *args: Any, **kwargs: Any
    ) -> Any:
        inner_function = ast.parse(self.api_code).body[0]
        inner_function.decorator_list = []
        # compile the function
        raw_byte_code = compile(ast.unparse(inner_function), "<string>", "exec")

        # load it
        exec(raw_byte_code)  # nosec

        internal_context = self.build_internal_context(context=context)

        # execute it
        evil_string = f"{self.func_name}(*args, **kwargs,context=internal_context)"
        result = eval(evil_string, None, locals())  # nosec

        # Update code context state
        self.update_state(internal_context.state)

        # return the results
        return result


@serializable()
class PrivateAPIEndpoint(Endpoint):
    # version
    __canonical_name__ = "PrivateAPIEndpoint"
    __version__ = SYFT_OBJECT_VERSION_1

    view_access: bool = False


@serializable()
class PublicAPIEndpoint(Endpoint):
    # version
    __canonical_name__ = "PublicAPIEndpoint"
    __version__ = SYFT_OBJECT_VERSION_1


class BaseTwinAPIEndpoint(SyftObject):
    __canonical_name__ = "BaseTwinAPIEndpoint"
    __version__ = SYFT_OBJECT_VERSION_1

    @model_validator(mode="before")
    @classmethod
    def validate_signature(cls, data: dict[str, Any]) -> dict[str, Any]:
        mock_function = data["mock_function"]  # mock_function can't be None
        private_function = data.get("private_function")

        # Add none check
        if private_function and private_function.signature != mock_function.signature:
            raise ValueError(
                "Mock and Private API Endpoints must have the same signature."
            )

        return data

    @field_validator("path", check_fields=False)
    @classmethod
    def validate_path(cls, path: str) -> str:
        # TODO: Check path doesn't collide with system endpoints

        if path.startswith(".") or path.endswith("."):
            raise ValueError("Path cannot start or end with a '.'")
        if not path.islower():
            raise ValueError("Path must be lowercase")
        parts = path.split(".")
        for part in parts:
            if not str.isidentifier(part) or keyword.iskeyword(part):
                raise ValueError(f"Invalid path: {part} is not a valid identifier")

        return path

    @field_validator("private_function", check_fields=False)
    @classmethod
    def validate_private_function(
        cls, private_function: PrivateAPIEndpoint | None
    ) -> PrivateAPIEndpoint | None:
        # TODO: what kind of validation should we do here?

        return private_function

    @field_validator("mock_function", check_fields=False)
    @classmethod
    def validate_mock_function(
        cls, mock_function: PublicAPIEndpoint
    ) -> PublicAPIEndpoint:
        # TODO: what kind of validation should we do here?
        return mock_function


@serializable()
class UpdateTwinAPIEndpoint(PartialSyftObject, BaseTwinAPIEndpoint):
    # version
    __canonical_name__ = "UpdateTwinAPIEndpoint"
    __version__ = SYFT_OBJECT_VERSION_1

    path: str
    private_function: PrivateAPIEndpoint | None = None
    mock_function: PublicAPIEndpoint
    description: MarkdownDescription | None = None
    endpoint_timeout: int = 60


@serializable()
class CreateTwinAPIEndpoint(BaseTwinAPIEndpoint):
    # version
    __canonical_name__ = "CreateTwinAPIEndpoint"
    __version__ = SYFT_OBJECT_VERSION_1

    path: str
    private_function: PrivateAPIEndpoint | None = None
    mock_function: PublicAPIEndpoint
    signature: Signature
    description: MarkdownDescription | None = None
    worker_pool: str | None = None
    endpoint_timeout: int = 60

    def __init__(
        self, description: str | MarkdownDescription | None = "", **kwargs: Any
    ) -> None:
        if isinstance(description, str):
            description = MarkdownDescription(text=description)

        super().__init__(**kwargs, description=description)


@serializable()
class TwinAPIEndpoint(SyncableSyftObject):
    # version
    __canonical_name__: str = "TwinAPIEndpoint"
    __version__ = SYFT_OBJECT_VERSION_1
    __exclude_sync_diff_attrs__ = ["private_function"]
    __private_sync_attr_mocks__ = {
        "private_function": None,
    }

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

    path: str
    private_function: PrivateAPIEndpoint | None = None
    mock_function: PublicAPIEndpoint
    signature: Signature
    description: MarkdownDescription | None = None
    action_object_id: UID
    worker_pool: str | None = None
    endpoint_timeout: int = 60

    __attr_searchable__ = ["path"]
    __attr_unique__ = ["path"]
    __repr_attrs__ = [
        "path",
        "description",
        "private_function",
        "mock_function",
        "endpoint_timeout",
    ]

    def has_mock(self) -> bool:
        return self.api_mock_code is not None

    def has_permission(self, context: AuthedServiceContext) -> bool:
        """Check if the user has permission to access the endpoint.

        Args:
            context: The context of the user requesting the code.
        Returns:
            bool: True if the user has permission to access the endpoint, False otherwise.
        """
        if context.role.value == 128:
            return True
        return False

    def select_code(self, context: AuthedServiceContext) -> Result[Ok, Err]:
        """Select the code to execute based on the user's permissions and public code availability.

        Args:
            context: The context of the user requesting the code.
        Returns:
            Result[Ok, Err]: The selected code to execute.
        """
        if self.has_permission(context) and self.private_function:
            return Ok(self.private_function)
        return Ok(self.mock_function)

    def exec(self, context: AuthedServiceContext, *args: Any, **kwargs: Any) -> Any:
        """Execute the code based on the user's permissions and public code availability.

        Args:
            context: The context of the user requesting the code.
            *args: Any
            **kwargs: Any
        Returns:
            Any: The result of the executed code.
        """
        result = self.select_code(context)
        if result.is_err():
            return SyftError(message=result.err())

        selected_code = result.ok()
        return self.exec_code(selected_code, context, *args, **kwargs)

    def exec_mock_function(
        self, context: AuthedServiceContext, *args: Any, **kwargs: Any
    ) -> Any:
        """Execute the public code if it exists."""
        if self.mock_function:
            return self.exec_code(self.mock_function, context, *args, **kwargs)

        return SyftError(message="No public code available")

    def exec_private_function(
        self, context: AuthedServiceContext, *args: Any, **kwargs: Any
    ) -> Any:
        """Execute the private code if user is has the proper permissions.

        Args:
            context: The context of the user requesting the code.
            *args: Any
            **kwargs: Any
        Returns:
            Any: The result of the executed code.
        """
        if self.private_function is None:
            return SyftError(message="No private code available")

        if self.has_permission(context):
            return self.exec_code(self.private_function, context, *args, **kwargs)

        return SyftError(message="You're not allowed to run this code.")

    def get_user_client_from_server(self, context: AuthedServiceContext) -> SyftClient:
        # get a user client
        guest_client = context.server.get_guest_client()
        user_client = guest_client
        signing_key_for_verify_key = context.server.get_service_method(
            UserService.signing_key_for_verify_key
        )
        private_key = signing_key_for_verify_key(
            context=context, verify_key=context.credentials
        )
        signing_key = private_key.signing_key
        user_client.credentials = signing_key
        return user_client

    def get_admin_client_from_server(self, context: AuthedServiceContext) -> SyftClient:
        admin_client = context.server.get_guest_client()
        admin_client.credentials = context.server.signing_key
        return admin_client

    def exec_code(
        self,
        code: PrivateAPIEndpoint | PublicAPIEndpoint,
        context: AuthedServiceContext,
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        try:
            inner_function = ast.parse(code.api_code).body[0]
            inner_function.decorator_list = []
            # compile the function
            src = ast.unparse(inner_function)
            raw_byte_code = compile(src, code.func_name, "exec")
            register_fn_in_linecache(code.func_name, src)
            user_client = self.get_user_client_from_server(context)
            admin_client = self.get_admin_client_from_server(context)

            # load it
            exec(raw_byte_code)  # nosec

            internal_context = code.build_internal_context(
                context=context, admin_client=admin_client, user_client=user_client
            )

            # execute it
            evil_string = f"{code.func_name}(*args, **kwargs,context=internal_context)"
            result = eval(evil_string, None, locals())  # nosec

            # Update code context state
            code.update_state(internal_context.state)

            if isinstance(code, PublicAPIEndpoint):
                self.mock_function = code
            else:
                self.private_function = code  # type: ignore

            api_service = context.server.get_service("apiservice")
            upsert_result = api_service.stash.upsert(
                context.server.get_service("userservice").admin_verify_key(), self
            )

            if upsert_result.is_err():
                raise Exception(upsert_result.err())

            # return the results
            return result
        except Exception as e:
            # If it's admin, return the error message.
            # TODO: cleanup typeerrors
            if context.role.value == 128 or isinstance(e, TypeError):
                return SyftError(
                    message=f"An error was raised during the execution of the API endpoint call: \n {str(e)}"
                )
            else:
                return SyftError(
                    message="Ops something went wrong during this endpoint execution, please contact your admin."
                )


def set_access_type(context: TransformContext) -> TransformContext:
    if context.output is not None and context.obj is not None:
        if context.obj.private_function is not None:
            context.output["access"] = "Private / Mock"
        else:
            context.output["access"] = "Public"
    return context


def check_and_cleanup_signature(context: TransformContext) -> TransformContext:
    if context.output is not None and context.obj is not None:
        params = dict(context.obj.signature.parameters)
        if "context" not in params:
            raise ValueError(
                "Function Signature must include 'context' [AuthedContext] parameters."
            )
        params.pop("context", None)
        context.output["signature"] = Signature(
            list(params.values()),
            return_annotation=context.obj.signature.return_annotation,
        )
    return context


def decorator_cleanup(code: str) -> str:
    # Regular expression to remove decorator
    # It matches from "@" to "def" (non-greedy) across multiple lines
    decorator_regex = r"@.*?def "

    # Substituting the matched pattern with "def"
    return re.sub(decorator_regex, "def ", code, count=1, flags=re.DOTALL)


def extract_code_string(code_field: str) -> Callable:
    def code_string(context: TransformContext) -> TransformContext:
        if context.obj is not None and context.output is not None:
            endpoint_type = (
                context.obj.private_function
                if code_field == "private_function"
                else context.obj.mock_function
            )
            helper_function_field = (
                "mock_helper_functions"
                if code_field == "mock_function"
                else "private_helper_functions"
            )

            context.server = cast(AbstractServer, context.server)
            admin_key = context.server.get_service("userservice").admin_verify_key()

            # If endpoint exists **AND** (has visible access **OR** the user is admin)
            if endpoint_type is not None and (
                endpoint_type.view_access or context.credentials == admin_key
            ):
                context.output[code_field] = decorator_cleanup(endpoint_type.api_code)
                context.output[helper_function_field] = (
                    endpoint_type.helper_functions.values() or []
                )
            else:
                context.output[code_field] = None
                context.output[helper_function_field] = []
        return context

    return code_string


@transform(TwinAPIAuthedContext, TwinAPIContextView)
def twin_api_context_to_twin_api_context_view() -> list[Callable]:
    return [keep(["state", "settings", "user"])]


@transform(CreateTwinAPIEndpoint, TwinAPIEndpoint)
def endpoint_create_to_twin_endpoint() -> list[Callable]:
    return [generate_id, generate_action_object_id, check_and_cleanup_signature]


@transform(TwinAPIEndpoint, TwinAPIEndpointView)
def twin_endpoint_to_view() -> list[Callable]:
    return [
        set_access_type,
        extract_code_string("private_function"),
        extract_code_string("mock_function"),
    ]


@transform(Endpoint, PrivateAPIEndpoint)
def endpoint_to_private_endpoint() -> list[Callable]:
    return [
        keep(
            [
                "api_code",
                "func_name",
                "settings",
                "view_access",
                "helper_functions",
                "state",
                "signature",
            ]
        )
    ]


@transform(Endpoint, PublicAPIEndpoint)
def endpoint_to_public_endpoint() -> list[Callable]:
    return [
        keep(
            [
                "api_code",
                "func_name",
                "settings",
                "view_access",
                "helper_functions",
                "state",
                "signature",
            ]
        )
    ]


def api_endpoint(
    path: str,
    settings: dict[str, str] | None = None,
    helper_functions: list[Callable] | None = None,
    description: MarkdownDescription | None = None,
    worker_pool: str | None = None,
    endpoint_timeout: int = 60,
) -> Callable[..., TwinAPIEndpoint | SyftError]:
    def decorator(f: Callable) -> TwinAPIEndpoint | SyftError:
        try:
            helper_functions_dict = {
                f.__name__: inspect.getsource(f) for f in (helper_functions or [])
            }
            res = CreateTwinAPIEndpoint(
                path=path,
                mock_function=PublicAPIEndpoint(
                    api_code=inspect.getsource(f),
                    func_name=f.__name__,
                    settings=settings,
                    signature=inspect.signature(f),
                    helper_functions=helper_functions_dict,
                ),
                signature=inspect.signature(f),
                description=description,
                worker_pool=worker_pool,
                endpoint_timeout=endpoint_timeout,
            )
        except ValidationError as e:
            for error in e.errors():
                error_msg = error["msg"]
            res = SyftError(message=error_msg)
        return res

    return decorator


def api_endpoint_method(
    settings: dict[str, str] | None = None,
    helper_functions: list[Callable] | None = None,
) -> Callable[..., Endpoint | SyftError]:
    def decorator(f: Callable) -> Endpoint | SyftError:
        try:
            helper_functions_dict = {
                f.__name__: inspect.getsource(f) for f in (helper_functions or [])
            }
            return Endpoint(
                api_code=inspect.getsource(f),
                func_name=f.__name__,
                settings=settings,
                signature=inspect.signature(f),
                helper_functions=helper_functions_dict,
            )
        except ValidationError as e:
            for error in e.errors():
                error_msg = error["msg"]
            res = SyftError(message=error_msg)
        return res

    return decorator


def create_new_api_endpoint(
    path: str,
    mock_function: Endpoint,
    private_function: Endpoint | None = None,
    description: MarkdownDescription | None = None,
    worker_pool: str | None = None,
    endpoint_timeout: int = 60,
    hide_mock_definition: bool = False,
    hide_private_definition: bool = True,
) -> CreateTwinAPIEndpoint | SyftError:
    try:
        # Parse the string to extract the function name

        endpoint_signature = mock_function.signature
        if private_function is not None:
            if private_function.signature != mock_function.signature:
                return SyftError(message="Signatures don't match")
            endpoint_signature = mock_function.signature
            private_function.view_access = not hide_private_definition
            mock_function.view_access = not hide_mock_definition

            return CreateTwinAPIEndpoint(
                path=path,
                private_function=private_function.to(PrivateAPIEndpoint),
                mock_function=mock_function.to(PublicAPIEndpoint),
                signature=endpoint_signature,
                description=description,
                worker_pool=worker_pool,
                endpoint_timeout=endpoint_timeout,
            )

        return CreateTwinAPIEndpoint(
            path=path,
            prublic_code=mock_function.to(PublicAPIEndpoint),
            signature=endpoint_signature,
            worker_pool=worker_pool,
            endpoint_timeout=endpoint_timeout,
        )
    except ValidationError as e:
        for error in e.errors():
            error_msg = error["msg"]

    return SyftError(message=error_msg)



# stdlib
from collections import defaultdict
from collections.abc import Iterable
from collections.abc import MutableMapping
from collections.abc import MutableSequence
from functools import cache
import hashlib
import json
from operator import itemgetter
import os
from pathlib import Path
import re
from types import UnionType
import typing
from typing import Any
import warnings

# third party
from packaging.version import parse
from result import OkErr
from result import Result

# relative
from .. import __version__
from ..service.response import SyftError
from ..service.response import SyftException
from ..service.response import SyftSuccess
from ..types.dicttuple import DictTuple
from ..types.syft_object import SyftBaseObject
from ..types.syft_object_registry import SyftObjectRegistry
from ..util.util import get_dev_mode

PROTOCOL_STATE_FILENAME = "protocol_version.json"
PROTOCOL_TYPE = str | int

IGNORE_TYPES = [
    "mock_type",
    "MockStore",
    "MockSyftObject",
    "MockStoreConfig",
    "MockWrapper",
    "base_stash_mock_object_type",
    "MockKeyValueBackingStore",
    "MockObjectFromSyftBaseObj",
    "MockObjectToSyftBaseObj",
]


def natural_key(key: PROTOCOL_TYPE) -> list[int | str | Any]:
    """Define key for natural ordering of strings."""
    if isinstance(key, int):
        key = str(key)
    return [int(s) if s.isdigit() else s for s in re.split(r"(\d+)", key)]


def sort_dict_naturally(d: dict) -> dict:
    """Sort dictionary by keys in natural order."""
    return {k: d[k] for k in sorted(d.keys(), key=natural_key)}


def data_protocol_file_name() -> str:
    return PROTOCOL_STATE_FILENAME


def data_protocol_dir() -> Path:
    return Path(os.path.abspath(str(Path(__file__).parent)))


def protocol_release_dir() -> Path:
    return data_protocol_dir() / "releases"


def handle_union_type_klass_name(type_klass_name: str) -> str:
    if type_klass_name == typing.Union.__name__:
        return UnionType.__name__
    return type_klass_name


def handle_annotation_repr_(annotation: type) -> str:
    """Handle typing representation."""
    origin = typing.get_origin(annotation)
    args = typing.get_args(annotation)
    if origin and args:
        args_repr = ", ".join(getattr(arg, "__name__", str(arg)) for arg in args)
        origin_repr = getattr(origin, "__name__", str(origin))

        # Handle typing.Union and types.UnionType
        origin_repr = handle_union_type_klass_name(origin_repr)
        return f"{origin_repr}: [{args_repr}]"
    elif args:
        args_repr = ", ".join(
            getattr(arg, "__name__", str(arg)) for arg in sorted(args)
        )
        return args_repr
    else:
        return repr(annotation)


class DataProtocol:
    def __init__(self, filename: str, raise_exception: bool = False) -> None:
        self.file_path = data_protocol_dir() / filename
        self.raise_exception = raise_exception
        self.load_state()

    def load_state(self) -> None:
        self.protocol_history = self.read_history()
        self.state = self.build_state()
        self.diff, self.current = self.diff_state(self.state)
        self.protocol_support = self.calculate_supported_protocols()

    @staticmethod
    def _calculate_object_hash(klass: type[SyftBaseObject]) -> str:
        # TODO: this depends on what is marked as serde

        # Rebuild the model to ensure that the fields are up to date
        # and any ForwardRef are resolved
        klass.model_rebuild()
        field_data = {
            field: handle_annotation_repr_(field_info.rebuild_annotation())
            for field, field_info in sorted(
                klass.model_fields.items(), key=itemgetter(0)
            )
        }
        obj_meta_info = {
            "canonical_name": klass.__canonical_name__,
            "version": klass.__version__,
            "unique_keys": getattr(klass, "__attr_unique__", []),
            "field_data": field_data,
        }

        return hashlib.sha256(json.dumps(obj_meta_info).encode()).hexdigest()

    @staticmethod
    def read_json(file_path: Path) -> dict:
        try:
            return json.loads(file_path.read_text())
        except Exception:
            return {}

    def read_history(self) -> dict:
        protocol_history = self.read_json(self.file_path)

        for version in protocol_history.keys():
            if version == "dev":
                continue
            release_version_path = (
                protocol_release_dir() / protocol_history[version]["release_name"]
            )
            released_version = self.read_json(file_path=release_version_path)
            protocol_history[version] = released_version.get(version, {})

        return protocol_history

    def save_history(self, history: dict) -> None:
        if os.path.exists(protocol_release_dir()):
            for file_path in protocol_release_dir().iterdir():
                for version in self.read_json(file_path):
                    # Skip adding file if the version is not part of the history
                    if version not in history.keys():
                        continue
                    history[version] = {"release_name": file_path.name}
        self.file_path.write_text(json.dumps(history, indent=2) + "\n")

    @property
    def latest_version(self) -> PROTOCOL_TYPE:
        sorted_versions = sorted(self.protocol_history.keys(), key=natural_key)
        if len(sorted_versions) > 0:
            return sorted_versions[-1] if self.has_dev else int(sorted_versions[-1])
        return "dev"

    @staticmethod
    def _hash_to_sha256(obj_dict: dict) -> str:
        return hashlib.sha256(json.dumps(obj_dict).encode()).hexdigest()

    def build_state(self, stop_key: str | None = None) -> dict:
        sorted_dict = sort_dict_naturally(self.protocol_history)
        state_dict: dict = defaultdict(dict)
        for protocol_number in sorted_dict:
            object_versions = sorted_dict[protocol_number]["object_versions"]
            for canonical_name, versions in object_versions.items():
                for version, object_metadata in versions.items():
                    action = object_metadata["action"]
                    version = object_metadata["version"]
                    hash_str = object_metadata["hash"]
                    state_versions = state_dict[canonical_name]
                    state_version_hashes = [val[0] for val in state_versions.values()]
                    if action == "add" and (
                        str(version) in state_versions.keys()
                        or hash_str in state_version_hashes
                    ):
                        raise Exception(
                            f"Can't add {object_metadata} already in state {versions}"
                        )
                    if action == "remove" and (
                        str(version) not in state_versions.keys()
                        and hash_str not in state_version_hashes
                    ):
                        raise Exception(
                            f"Can't remove {object_metadata} missing from state {versions} for object {canonical_name}."
                        )
                    if action == "add":
                        state_dict[canonical_name][str(version)] = (
                            hash_str,
                            protocol_number,
                        )
                    elif action == "remove":
                        del state_dict[canonical_name][str(version)]
            # stop early
            if stop_key == protocol_number:
                return state_dict
        return state_dict

    @staticmethod
    def obj_json(version: str | int, _hash: str, action: str = "add") -> dict:
        return {
            "version": int(version),
            "hash": _hash,
            "action": action,
        }

    def diff_state(self, state: dict) -> tuple[dict, dict]:
        compare_dict: dict = defaultdict(dict)  # what versions are in the latest code
        object_diff: dict = defaultdict(dict)  # diff in latest code with saved json
        all_serde_propeties = [
            serde_properties
            for version_dict in SyftObjectRegistry.__object_serialization_registry__.values()
            for serde_properties in version_dict.values()
        ]
        for serde_properties in all_serde_propeties:
            cls, version = serde_properties[7], serde_properties[9]
            if issubclass(cls, SyftBaseObject):
                canonical_name = cls.__canonical_name__
                if canonical_name in IGNORE_TYPES or canonical_name.startswith(
                    "MockSyftObject_"
                ):
                    continue

                hash_str = DataProtocol._calculate_object_hash(cls)

                # build this up for later
                compare_dict[canonical_name][str(version)] = hash_str

                if canonical_name not in state:
                    # new object so its an add
                    obj_to_add = self.obj_json(int(version), hash_str)
                    object_diff[canonical_name][str(version)] = obj_to_add
                    continue

                versions = state[canonical_name]
                if (
                    str(version) in versions.keys()
                    and versions[str(version)][0] == hash_str
                ):
                    # already there so do nothing
                    continue
                elif str(version) in versions.keys():
                    is_protocol_dev = versions[str(version)][1] == "dev"
                    if is_protocol_dev:
                        # force overwrite existing object so its an add
                        obj_to_add = self.obj_json(int(version), hash_str)
                        object_diff[canonical_name][str(version)] = obj_to_add
                        continue
                    error_msg = f"""{canonical_name} for class {cls.__name__} fqn {cls}\
version {version} hash has changed. {hash_str} not in {versions.values()}. \
Is a unique __canonical_name__ for this subclass missing?
If the class has changed you will need to define a new class with the changes, \
with same __canonical_name__ and bump the __version__ number. {cls.model_fields}
"""

                    if get_dev_mode() or self.raise_exception:
                        raise Exception(error_msg)
                    else:
                        warnings.warn(error_msg, stacklevel=1, category=UserWarning)
                        break
                else:
                    # new object so its an add
                    obj_to_add = self.obj_json(int(version), hash_str)
                    object_diff[canonical_name][str(version)] = obj_to_add
                    continue

        # now check for remove actions
        for canonical_name in state:
            for version, (hash_str, _) in state[canonical_name].items():
                if canonical_name not in compare_dict:
                    # missing so its a remove
                    obj_to_remove = self.obj_json(int(version), hash_str, "remove")
                    object_diff[canonical_name][str(version)] = obj_to_remove
                    continue
                versions = compare_dict[canonical_name]
                if str(version) not in versions.keys():
                    # missing so its a remove
                    obj_to_remove = self.obj_json(int(version), hash_str, "remove")
                    object_diff[canonical_name][str(version)] = obj_to_remove
                    continue
        return object_diff, compare_dict

    def stage_protocol_changes(self) -> Result[SyftSuccess, SyftError]:
        change_count = 0
        current_history = self.protocol_history
        if "dev" not in current_history:
            current_history["dev"] = {}
            current_history["dev"]["object_versions"] = {}
        object_versions = current_history["dev"]["object_versions"]
        for canonical_name, versions in self.diff.items():
            for version, version_metadata in versions.items():
                if canonical_name not in object_versions:
                    object_versions[canonical_name] = {}
                change_count += 1
                action = version_metadata["action"]

                # Allow removal of class that only been staged to dev
                if (
                    action == "remove"
                    and str(version) in object_versions[canonical_name]
                ):
                    # Delete the whole class if only single version exists
                    if len(object_versions[canonical_name]) == 1:
                        del object_versions[canonical_name]
                    else:
                        # In case of multiple versions of the class only delete the selected
                        del object_versions[canonical_name][str(version)]

                else:  # Add or overwrite existing data in dev
                    object_versions[canonical_name][str(version)] = version_metadata

            # Sort the version dict
            object_versions[canonical_name] = sort_dict_naturally(
                object_versions.get(canonical_name, {})
            )

        current_history["dev"]["object_versions"] = object_versions

        # trim empty dev
        if len(current_history["dev"]["object_versions"]) == 0:
            del current_history["dev"]

        self.save_history(current_history)
        self.load_state()
        return SyftSuccess(message=f"{change_count} Protocol Updates Staged to dev")

    def bump_protocol_version(self) -> Result[SyftSuccess, SyftError]:
        if len(self.diff):
            raise Exception(
                "You can't bump the protocol version with unstaged changes."
            )

        keys = self.protocol_history.keys()
        if "dev" not in keys:
            self.validate_release()
            print("You can't bump the protocol if there are no staged changes.")
            return SyftError(
                message="Failed to bump version as there are no staged changes."
            )

        highest_protocol = 0
        for k in self.protocol_history.keys():
            if k == "dev":
                continue
            highest_protocol = max(highest_protocol, int(k))

        next_highest_protocol = highest_protocol + 1
        self.protocol_history[str(next_highest_protocol)] = self.protocol_history["dev"]
        self.freeze_release(self.protocol_history, str(next_highest_protocol))
        del self.protocol_history["dev"]
        self.save_history(self.protocol_history)
        self.load_state()
        return SyftSuccess(message=f"Protocol Updated to {next_highest_protocol}")

    @staticmethod
    def freeze_release(protocol_history: dict, latest_protocol: str) -> None:
        """Freezes latest release as a separate release file."""

        # Get release history
        release_history = protocol_history[latest_protocol]

        # Create new file for the version
        syft_version = parse(__version__)
        release_file_name = f"{syft_version.public}.json"
        release_file = protocol_release_dir() / release_file_name

        # Save the new released version
        release_file.write_text(
            json.dumps({latest_protocol: release_history}, indent=2)
        )

    def validate_release(self) -> None:
        """Validate if latest release name is consistent with syft version"""
        # Read the protocol history
        protocol_history = self.read_json(self.file_path)
        sorted_protocol_versions = sorted(protocol_history.keys(), key=natural_key)

        # Grab the latest protocol
        latest_protocol = (
            sorted_protocol_versions[-1] if len(sorted_protocol_versions) > 0 else None
        )

        # Skip validation if latest protocol is dev
        if latest_protocol is None or latest_protocol == "dev":
            return

        # Get filename of the latest protocol
        release_name = protocol_history[latest_protocol]["release_name"]
        # Extract syft version from release name
        protocol_syft_version = parse(release_name.split(".json")[0])
        current_syft_version = parse(__version__)

        # If base syft version in latest protocol version is not same as current syft version
        # Skip updating the release name
        if protocol_syft_version.base_version != current_syft_version.base_version:
            return

        # Update release name to latest beta, stable or post based on current syft version
        print(
            f"Current release {release_name} will be updated to {current_syft_version}"
        )

        # Get latest protocol file path
        latest_protocol_fp: Path = protocol_release_dir() / release_name

        # New protocol file path
        new_protocol_file_path = (
            protocol_release_dir() / f"{current_syft_version.public}.json"
        )

        # Update older file path to newer file path
        latest_protocol_fp.rename(new_protocol_file_path)
        protocol_history[latest_protocol]["release_name"] = (
            f"{current_syft_version}.json"
        )

        # Save history
        self.file_path.write_text(json.dumps(protocol_history, indent=2) + "\n")

        # Reload protocol
        self.read_history()

    def revert_latest_protocol(self) -> Result[SyftSuccess, SyftError]:
        """Revert latest protocol changes to dev"""

        # Get current protocol history
        protocol_history = self.read_json(self.file_path)

        # Get latest released protocol
        sorted_protocol_versions = sorted(protocol_history.keys(), key=natural_key)
        latest_protocol = (
            sorted_protocol_versions[-1] if len(sorted_protocol_versions) > 0 else None
        )

        # If current protocol is dev, skip revert
        if latest_protocol is None or latest_protocol == "dev":
            return SyftError(message="Revert skipped !! Already running dev protocol.")

        # Read the current released protocol
        release_name = protocol_history[latest_protocol]["release_name"]
        protocol_file_path: Path = protocol_release_dir() / release_name

        released_protocol = self.read_json(protocol_file_path)
        protocol_history["dev"] = released_protocol[latest_protocol]

        # Delete the current released protocol
        protocol_history.pop(latest_protocol)
        protocol_file_path.unlink()

        # Save history
        self.save_history(protocol_history)
        self.load_state()

    def check_protocol(self) -> Result[SyftSuccess, SyftError]:
        if len(self.diff) != 0:
            return SyftError(message="Protocol Changes Unstaged")
        else:
            return SyftSuccess(message="Protocol Stable")

    def check_or_stage_protocol(self) -> Result[SyftSuccess, SyftError]:
        if not self.check_protocol():
            self.stage_protocol_changes()
        result = self.check_protocol()
        return result

    @property
    def supported_protocols(self) -> list[int | str]:
        """Returns a list of protocol numbers that are marked as supported."""
        supported = []
        for version, is_supported in self.protocol_support.items():
            if is_supported:
                if version != "dev":
                    version = int(version)
                supported.append(version)
        return supported

    def calculate_supported_protocols(self) -> dict:
        protocol_supported = {}
        # go through each historical protocol version
        for v, version_data in self.protocol_history.items():
            # we assume its supported until we prove otherwise
            protocol_supported[v] = True
            # iterate through each object
            for canonical_name in version_data["object_versions"].keys():
                if canonical_name not in self.state:
                    protocol_supported[v] = False
                    break
        return protocol_supported

    def get_object_versions(self, protocol: int | str) -> list:
        return self.protocol_history[str(protocol)]["object_versions"]

    @property
    def has_dev(self) -> bool:
        if "dev" in self.protocol_history.keys():
            return True
        return False

    def reset_dev_protocol(self) -> None:
        if self.has_dev:
            del self.protocol_history["dev"]
            self.save_history(self.protocol_history)


def get_data_protocol(raise_exception: bool = False) -> DataProtocol:
    return _get_data_protocol(
        filename=data_protocol_file_name(),
        raise_exception=raise_exception,
    )


@cache
def _get_data_protocol(filename: str, raise_exception: bool = False) -> DataProtocol:
    return DataProtocol(
        filename=filename,
        raise_exception=raise_exception,
    )


def stage_protocol_changes() -> Result[SyftSuccess, SyftError]:
    data_protocol = get_data_protocol(raise_exception=True)
    return data_protocol.stage_protocol_changes()


def bump_protocol_version() -> Result[SyftSuccess, SyftError]:
    data_protocol = get_data_protocol(raise_exception=True)
    return data_protocol.bump_protocol_version()


def check_or_stage_protocol() -> Result[SyftSuccess, SyftError]:
    data_protocol = get_data_protocol()
    return data_protocol.check_or_stage_protocol()


def debox_arg_and_migrate(arg: Any, protocol_state: dict) -> Any:
    """Debox the argument based on whether it is iterable or single entity."""
    constructor = None
    extra_args = []

    single_entity = False

    if isinstance(arg, OkErr):
        constructor = type(arg)
        arg = arg.value

    if isinstance(arg, MutableMapping):
        iterable_keys: Iterable = arg.keys()
    elif isinstance(arg, MutableSequence):
        iterable_keys = range(len(arg))
    elif isinstance(arg, tuple):
        iterable_keys = range(len(arg))
        constructor = type(arg)
        if isinstance(arg, DictTuple):
            extra_args.append(arg.keys())
        arg = list(arg)
    else:
        iterable_keys = range(1)
        arg = [arg]
        single_entity = True

    for key in iterable_keys:
        _object = arg[key]
        if isinstance(_object, SyftBaseObject):
            current_version = int(_object.__version__)
            migrate_to_version = int(max(protocol_state[_object.__canonical_name__]))
            if current_version > migrate_to_version:  # downgrade
                versions = range(current_version - 1, migrate_to_version - 1, -1)
            else:  # upgrade
                versions = range(current_version + 1, migrate_to_version + 1)
            for version in versions:
                _object = _object.migrate_to(version)
        arg[key] = _object

    wrapped_arg = arg[0] if single_entity else arg
    if constructor is not None:
        wrapped_arg = constructor(wrapped_arg, *extra_args)

    return wrapped_arg


def migrate_args_and_kwargs(
    args: tuple,
    kwargs: dict,
    to_protocol: PROTOCOL_TYPE | None = None,
    to_latest_protocol: bool = False,
) -> tuple[tuple, dict]:
    """Migrate args and kwargs to latest version for given protocol.

    If `to_protocol` is None, then migrate to latest protocol version.

    """
    data_protocol = get_data_protocol()

    if to_protocol is None:
        to_protocol = data_protocol.latest_version if to_latest_protocol else None

    if to_protocol is None:
        raise SyftException("Protocol version missing.")

    # If latest protocol being used is equal to the protocol to be migrate
    # then skip migration of the object
    if to_protocol == data_protocol.latest_version:
        return args, kwargs

    protocol_state = data_protocol.build_state(stop_key=str(to_protocol))

    migrated_kwargs, migrated_args = {}, []

    for param_name, param_val in kwargs.items():
        migrated_val = debox_arg_and_migrate(
            arg=param_val,
            protocol_state=protocol_state,
        )
        migrated_kwargs[param_name] = migrated_val

    for arg in args:
        migrated_val = debox_arg_and_migrate(
            arg=arg,
            protocol_state=protocol_state,
        )
        migrated_args.append(migrated_val)

    return tuple(migrated_args), migrated_kwargs



# stdlib

# third party
from result import Err
from result import Ok
from result import Result

# relative
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...service.response import SyftError
from ...store.document_store import BaseStash
from ...store.document_store import DocumentStore
from ...store.document_store import PartitionKey
from ...store.document_store import PartitionSettings
from ...types.uid import UID
from ...util.telemetry import instrument
from ..action.action_permissions import ActionObjectPermission
from .notifier import NotifierSettings

NamePartitionKey = PartitionKey(key="name", type_=str)
ActionIDsPartitionKey = PartitionKey(key="action_ids", type_=list[UID])


@instrument
@serializable(canonical_name="NotifierStash", version=1)
class NotifierStash(BaseStash):
    object_type = NotifierSettings
    settings: PartitionSettings = PartitionSettings(
        name=NotifierSettings.__canonical_name__, object_type=NotifierSettings
    )

    def __init__(self, store: DocumentStore) -> None:
        super().__init__(store=store)

    def admin_verify_key(self) -> SyftVerifyKey:
        return self.partition.root_verify_key

    # TODO: should this method behave like a singleton?
    def get(self, credentials: SyftVerifyKey) -> Result[NotifierSettings, Err]:
        """Get Settings"""
        result = self.get_all(credentials)
        if result.is_ok():
            settings = result.ok()
            if len(settings) == 0:
                return Ok(
                    None
                )  # TODO: Stash shouldn't be empty after init. Return Err instead?
            result = settings[
                0
            ]  # TODO: Should we check if theres more than one? => Report corruption
            return Ok(result)
        else:
            return Err(SyftError(message=result.err()))

    def set(
        self,
        credentials: SyftVerifyKey,
        settings: NotifierSettings,
        add_permissions: list[ActionObjectPermission] | None = None,
        add_storage_permission: bool = True,
        ignore_duplicates: bool = False,
    ) -> Result[NotifierSettings, Err]:
        result = self.check_type(settings, self.object_type)
        # we dont use and_then logic here as it is hard because of the order of the arguments
        if result.is_err():
            return Err(SyftError(message=result.err()))
        return super().set(
            credentials=credentials, obj=result.ok()
        )  # TODO check if result isInstance(Ok)

    def update(
        self,
        credentials: SyftVerifyKey,
        settings: NotifierSettings,
        has_permission: bool = False,
    ) -> Result[NotifierSettings, Err]:
        result = self.check_type(settings, self.object_type)
        # we dont use and_then logic here as it is hard because of the order of the arguments
        if result.is_err():
            return Err(SyftError(message=result.err()))
        return super().update(
            credentials=credentials, obj=result.ok()
        )  # TODO check if result isInstance(Ok)



# stdlib
from enum import Enum

# relative
from ..serde.serializable import serializable


@serializable(canonical_name="SyncDirection", version=1)
class SyncDirection(str, Enum):
    LOW_TO_HIGH = "low_to_high"
    HIGH_TO_LOW = "high_to_low"

    def to_sync_decision(self) -> "SyncDecision":
        if self == SyncDirection.LOW_TO_HIGH:
            return SyncDecision.LOW
        elif self == SyncDirection.HIGH_TO_LOW:
            return SyncDecision.HIGH
        else:
            raise ValueError("Invalid SyncDirection")


class SyncDecision(Enum):
    LOW = "low"
    HIGH = "high"
    SKIP = "skip"
    IGNORE = "ignore"



# stdlib
from collections.abc import Callable

# relative
from ...types.transforms import TransformContext


def set_from_server_to_key(server_attr: str, key: str) -> Callable:
    def extract_from_server(context: TransformContext) -> TransformContext:
        if context.output is not None:
            context.output[key] = getattr(context.server, server_attr)
        return context

    return extract_from_server



# stdlib
import base64
from functools import lru_cache
import importlib.resources

IMAGE_ASSETS = "syft.assets.img"
SVG_ASSETS = "syft.assets.svg"
CSS_ASSETS = "syft.assets.css"
JS_ASSETS = "syft.assets.js"


@lru_cache(maxsize=32)
def load_svg(fname: str) -> str:
    # TODO add resize support
    return importlib.resources.read_text(SVG_ASSETS, fname)


@lru_cache(maxsize=32)
def load_png_base64(fname: str) -> str:
    b = importlib.resources.read_binary(IMAGE_ASSETS, fname)
    res = base64.b64encode(b)
    return f"data:image/png;base64,{res.decode('utf-8')}"


@lru_cache(maxsize=32)
def load_css(fname: str) -> str:
    return importlib.resources.read_text(CSS_ASSETS, fname)


@lru_cache(maxsize=32)
def load_js(fname: str) -> str:
    return importlib.resources.read_text(JS_ASSETS, fname)



# stdlib
from datetime import datetime
from datetime import timedelta
from datetime import timezone
from enum import Enum
import random
from string import Template
from time import sleep
from typing import Any

# third party
from pydantic import Field
from pydantic import model_validator
from result import Err
from result import Ok
from result import Result
from typing_extensions import Self

# relative
from ...client.api import APIRegistry
from ...client.api import SyftAPICall
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...service.context import AuthedServiceContext
from ...service.worker.worker_pool import SyftWorker
from ...store.document_store import BaseUIDStoreStash
from ...store.document_store import DocumentStore
from ...store.document_store import PartitionKey
from ...store.document_store import PartitionSettings
from ...store.document_store import QueryKeys
from ...store.document_store import UIDPartitionKey
from ...types.datetime import DateTime
from ...types.datetime import format_timedelta
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ...types.syft_object import SyftObject
from ...types.syncable_object import SyncableSyftObject
from ...types.uid import UID
from ...util import options
from ...util.colors import SURFACE
from ...util.markdown import as_markdown_code
from ...util.telemetry import instrument
from ...util.util import prompt_warning_message
from ..action.action_object import Action
from ..action.action_object import ActionObject
from ..action.action_permissions import ActionObjectPermission
from ..log.log import SyftLog
from ..response import SyftError
from ..response import SyftNotReady
from ..response import SyftSuccess
from ..user.user import UserView
from .html_template import job_repr_template


@serializable(canonical_name="JobStatus", version=1)
class JobStatus(str, Enum):
    CREATED = "created"
    PROCESSING = "processing"
    ERRORED = "errored"
    COMPLETED = "completed"
    TERMINATING = "terminating"
    INTERRUPTED = "interrupted"


def center_content(text: Any) -> str:
    if isinstance(text, str):
        text = text.replace("\n", "<br>")
    center_div = f"""
    <div style="
        display: flex;
        justify-content: center;
        align-items: center; width: 100%; height: 100%;">
        {text}
    </div>
    """
    center_div = center_div.replace("\n", "")
    return center_div


@serializable(canonical_name="JobType", version=1)
class JobType(str, Enum):
    JOB = "job"
    TWINAPIJOB = "twinapijob"

    def __str__(self) -> str:
        return self.value


@serializable()
class Job(SyncableSyftObject):
    __canonical_name__ = "JobItem"
    __version__ = SYFT_OBJECT_VERSION_1

    id: UID
    server_uid: UID
    result: Any | None = None
    resolved: bool = False
    status: JobStatus = JobStatus.CREATED
    log_id: UID | None = None
    parent_job_id: UID | None = None
    n_iters: int | None = 0
    current_iter: int | None = None
    creation_time: str | None = Field(
        default_factory=lambda: str(datetime.now(tz=timezone.utc))
    )
    action: Action | None = None
    job_pid: int | None = None
    job_worker_id: UID | None = None
    updated_at: DateTime | None = None
    user_code_id: UID | None = None
    requested_by: UID | None = None
    job_type: JobType = JobType.JOB

    __attr_searchable__ = ["parent_job_id", "job_worker_id", "status", "user_code_id"]
    __repr_attrs__ = [
        "id",
        "result",
        "resolved",
        "progress",
        "creation_time",
        "user_code_name",
    ]
    __exclude_sync_diff_attrs__ = ["action", "server_uid"]
    __table_coll_widths__ = [
        "min-content",
        "auto",
        "auto",
        "auto",
        "auto",
        "auto",
        "auto",
    ]
    __syft_include_id_coll_repr__ = False

    @model_validator(mode="after")
    def check_user_code_id(self) -> Self:
        if self.action is not None:
            if self.user_code_id is None:
                self.user_code_id = self.action.user_code_id
            elif self.action.user_code_id != self.user_code_id:
                raise ValueError(
                    "user_code_id does not match the action's user_code_id",
                    self.__class__,
                )

        return self

    @property
    def action_display_name(self) -> str:
        if self.action is None:
            return "action"
        else:
            # hacky
            self.action.syft_server_location = self.syft_server_location
            self.action.syft_client_verify_key = self.syft_client_verify_key
            return self.action.job_display_name

    @property
    def user_code_name(self) -> str | None:
        if self.user_code_id is not None:
            api = APIRegistry.api_for(
                server_uid=self.syft_server_location,
                user_verify_key=self.syft_client_verify_key,
            )
            if api is None:
                return None
            user_code = api.services.code.get_by_id(self.user_code_id)
            return user_code.service_func_name
        return None

    @property
    def time_remaining_string(self) -> str | None:
        # update state
        self.fetch()
        if (
            self.current_iter is not None
            and self.n_iters is not None
            and self.n_iters != 0
        ):
            percentage = round((self.current_iter / self.n_iters) * 100)
            blocks_filled = round(percentage / 20)
            blocks_empty = 5 - blocks_filled
            blocks_filled_str = "█" * blocks_filled
            blocks_empty_str = "&nbsp;&nbsp;" * blocks_empty
            return f"{percentage}% |{blocks_filled_str}{blocks_empty_str}|\n{self.current_iter}/{self.n_iters}\n"
        return None

    @property
    def worker(self) -> SyftWorker | SyftError:
        api = APIRegistry.api_for(
            server_uid=self.syft_server_location,
            user_verify_key=self.syft_client_verify_key,
        )
        if api is None:
            return SyftError(
                message=f"Can't access Syft API. You must login to {self.syft_server_location}"
            )
        return api.services.worker.get(self.job_worker_id)

    @property
    def eta_string(self) -> str | None:
        if (
            self.current_iter is None
            or self.current_iter == 0
            or self.n_iters is None
            or self.creation_time is None
        ):
            return None

        now = datetime.now(tz=timezone.utc)
        time_passed = now - datetime.fromisoformat(self.creation_time)
        iter_duration_seconds: float = time_passed.total_seconds() / self.current_iter
        iters_remaining = self.n_iters - self.current_iter

        # TODO: Adjust by the number of consumers
        time_remaining = timedelta(seconds=iters_remaining * iter_duration_seconds)
        time_passed_str = format_timedelta(time_passed)
        time_remaining_str = format_timedelta(time_remaining)

        if iter_duration_seconds >= 1:
            iter_duration: timedelta = timedelta(seconds=iter_duration_seconds)
            iter_duration_str = f"{format_timedelta(iter_duration)}s/it"
        else:
            iters_per_second = round(1 / iter_duration_seconds)
            iter_duration_str = f"{iters_per_second}it/s"

        return f"[{time_passed_str}<{time_remaining_str}]\n{iter_duration_str}"

    @property
    def progress(self) -> str | None:
        if self.status in [JobStatus.PROCESSING, JobStatus.COMPLETED]:
            if self.current_iter is None:
                return ""
            else:
                if self.n_iters is not None:
                    return self.time_remaining_string
                # if self.current_iter !=0
                # we can compute the remaining time

                # we cannot compute the remaining time
                else:
                    n_iters_str = "?" if self.n_iters is None else str(self.n_iters)
                    return f"{self.current_iter}/{n_iters_str}"
        else:
            return ""

    def info(
        self,
        public_metadata: bool = True,
        result: bool = False,
    ) -> "JobInfo":
        return JobInfo.from_job(self, public_metadata, result)

    def apply_info(self, info: "JobInfo") -> None:
        if info.includes_metadata:
            for attr in info.__public_metadata_attrs__:
                setattr(self, attr, getattr(info, attr))

        if info.includes_result:
            self.result = info.result

    def restart(self, kill: bool = False) -> None:
        api = APIRegistry.api_for(
            server_uid=self.syft_server_location,
            user_verify_key=self.syft_client_verify_key,
        )
        if api is None:
            raise ValueError(
                f"Can't access Syft API. You must login to {self.syft_server_location}"
            )
        call = SyftAPICall(
            server_uid=self.server_uid,
            path="job.restart",
            args=[],
            kwargs={"uid": self.id},
            blocking=True,
        )
        res = api.make_call(call)
        self.fetch()
        return res

    def kill(self) -> SyftError | SyftSuccess:
        api = APIRegistry.api_for(
            server_uid=self.syft_server_location,
            user_verify_key=self.syft_client_verify_key,
        )
        if api is None:
            return SyftError(
                message=f"Can't access Syft API. You must login to {self.syft_server_location}"
            )
        call = SyftAPICall(
            server_uid=self.server_uid,
            path="job.kill",
            args=[],
            kwargs={"id": self.id},
            blocking=True,
        )
        res = api.make_call(call)
        self.fetch()
        return res

    def fetch(self) -> None:
        api = APIRegistry.api_for(
            server_uid=self.syft_server_location,
            user_verify_key=self.syft_client_verify_key,
        )
        if api is None:
            raise ValueError(
                f"Can't access Syft API. You must login to {self.syft_server_location}"
            )
        call = SyftAPICall(
            server_uid=self.server_uid,
            path="job.get",
            args=[],
            kwargs={"uid": self.id},
            blocking=True,
        )
        job: Job | None = api.make_call(call)
        if job is None:
            return None
        self.resolved = job.resolved
        if job.resolved:
            self.result = job.result

        self.status = job.status
        self.n_iters = job.n_iters
        self.current_iter = job.current_iter

    @property
    def subjobs(self) -> list["Job"] | SyftError:
        api = APIRegistry.api_for(
            server_uid=self.syft_server_location,
            user_verify_key=self.syft_client_verify_key,
        )
        if api is None:
            return SyftError(
                message=f"Can't access Syft API. You must login to {self.syft_server_location}"
            )
        return api.services.job.get_subjobs(self.id)

    def get_subjobs(self, context: AuthedServiceContext) -> list["Job"] | SyftError:
        job_service = context.server.get_service("jobservice")
        return job_service.get_subjobs(context, self.id)

    @property
    def owner(self) -> UserView | SyftError:
        api = APIRegistry.api_for(
            server_uid=self.syft_server_location,
            user_verify_key=self.syft_client_verify_key,
        )
        if api is None:
            return SyftError(
                message=f"Can't access Syft API. You must login to {self.syft_server_location}"
            )
        return api.services.user.get_current_user(self.id)

    def _get_log_objs(self) -> SyftLog | SyftError:
        api = APIRegistry.api_for(
            server_uid=self.server_uid,
            user_verify_key=self.syft_client_verify_key,
        )
        if api is None:
            raise ValueError(f"api is None. You must login to {self.server_uid}")
        return api.services.log.get(self.log_id)

    def logs(
        self, stdout: bool = True, stderr: bool = True, _print: bool = True
    ) -> str | None:
        api = APIRegistry.api_for(
            server_uid=self.syft_server_location,
            user_verify_key=self.syft_client_verify_key,
        )
        if api is None:
            return (
                f"Can't access Syft API. You must login to {self.syft_server_location}"
            )

        has_permissions = True

        results = []
        if stdout:
            stdout_log = api.services.log.get_stdout(self.log_id)
            if isinstance(stdout_log, SyftError):
                results.append(f"Log {self.log_id} not available")
                has_permissions = False
            else:
                results.append(stdout_log)

        if stderr:
            try:
                stderr_log = api.services.log.get_stderr(self.log_id)
                if isinstance(stderr_log, SyftError):
                    results.append(f"Error log {self.log_id} not available")
                    has_permissions = False
                else:
                    results.append(stderr_log)
            except Exception:
                # no access
                if isinstance(self.result, Err):
                    results.append(self.result.value)
        else:
            # add short error
            if isinstance(self.result, Err):
                results.append(self.result.value)

        if has_permissions:
            has_storage_permission = api.services.log.has_storage_permission(
                self.log_id
            )
            if not has_storage_permission:
                prompt_warning_message(
                    message="This is a placeholder object, the real data lives on a different server and is not synced."
                )

        results_str = "\n".join(results)
        if not _print:
            return results_str
        else:
            print(results_str)
            return None

    # def __repr__(self) -> str:
    #     return f"<Job: {self.id}>: {self.status}"

    def status_badge(self) -> dict[str, str]:
        status = self.status
        if status in [JobStatus.COMPLETED]:
            badge_color = "label-green"
        elif status in [JobStatus.PROCESSING]:
            badge_color = "label-orange"
        elif status in [JobStatus.CREATED]:
            badge_color = "label-gray"
        elif status in [JobStatus.ERRORED, JobStatus.INTERRUPTED]:
            badge_color = "label-red"
        else:
            badge_color = "label-orange"
        return {"value": status.upper(), "type": badge_color}

    def summary_html(self) -> str:
        # TODO: Fix id for buttons
        # relative
        from ...util.notebook_ui.components.sync import CopyIDButton

        try:
            # type_html = f'<div class="label {self.type_badge_class()}">{self.object_type_name.upper()}</div>'
            description_html = (
                f"<span class='syncstate-description'>{self.user_code_name}</span>"
            )
            worker_summary = ""
            if self.job_worker_id:
                worker_copy_button = CopyIDButton(
                    copy_text=str(self.job_worker_id), max_width=60
                )
                worker_summary = f"""
                <div style="display: table-row">
                    <span class='syncstate-col-footer'>{'on worker'}
                    {worker_copy_button.to_html()}</span>
                </div>
                """

            summary_html = f"""
                <div style="display: flex; gap: 8px; justify-content: start; width: 100%;">
                    {description_html}
                    <div style="display: flex; gap: 8px; justify-content: end; width: 100%;">
                        {CopyIDButton(copy_text=str(self.id), max_width=60).to_html()}
                    </div>
                </div>
                <div style="display: table-row">
                <span class='syncstate-col-footer'>{self.creation_time[:-7] if self.creation_time else ''}</span>
                </div>
                {worker_summary}
                """
            summary_html = summary_html.replace("\n", "")
        except Exception as e:
            print("Failed to build table", e)
            raise
        return summary_html

    def _coll_repr_(self) -> dict[str, Any]:
        logs = self.logs(_print=False, stderr=False)
        if logs is not None:
            log_lines = logs.split("\n")
        subjobs = self.subjobs
        if len(log_lines) > 2:
            logs = f"... ({len(log_lines)} lines)\n" + "\n".join(log_lines[-2:])

        def default_value(value: str) -> str:
            return value if value else "--"

        return {
            "Status": self.status_badge(),
            "Job": self.summary_html(),
            "# Subjobs": default_value(len(subjobs)),
            "Progress": default_value(self.progress),
            "ETA": default_value(self.eta_string),
            "Logs": default_value(logs),
        }

    @property
    def has_parent(self) -> bool:
        return self.parent_job_id is not None

    def _repr_markdown_(self, wrap_as_python: bool = True, indent: int = 0) -> str:
        _ = self.resolve
        logs = self.logs(_print=False)
        if logs is not None:
            logs_w_linenr = "\n".join(
                [f"{i} {line}" for i, line in enumerate(logs.rstrip().split("\n"))]
            )

        if self.status == JobStatus.COMPLETED:
            logs_w_linenr += "\nJOB COMPLETED"

        md = f"""class Job:
    id: UID = {self.id}
    status: {self.status}
    has_parent: {self.has_parent}
    result: {self.result.__str__()}
    logs:

{logs_w_linenr}
    """
        return as_markdown_code(md)

    @property
    def fetched_status(self) -> JobStatus:
        self.fetch()
        return self.status

    @property
    def requesting_user(self) -> UserView | SyftError:
        api = APIRegistry.api_for(
            server_uid=self.syft_server_location,
            user_verify_key=self.syft_client_verify_key,
        )
        if api is None:
            return SyftError(
                message=f"Can't access Syft API. You must login to {self.syft_server_location}"
            )
        return api.services.user.view(self.requested_by)

    @property
    def server_name(self) -> str | SyftError | None:
        api = APIRegistry.api_for(
            server_uid=self.syft_server_location,
            user_verify_key=self.syft_client_verify_key,
        )
        if api is None:
            return SyftError(
                message=f"Can't access Syft API. You must login to {self.syft_server_location}"
            )
        return api.server_name

    @property
    def parent(self) -> Self | SyftError:
        api = APIRegistry.api_for(
            server_uid=self.syft_server_location,
            user_verify_key=self.syft_client_verify_key,
        )
        if api is None:
            return SyftError(
                message=f"Can't access Syft API. You must login to {self.syft_server_location}"
            )
        return api.services.job.get(self.parent_job_id)

    @property
    def ancestors_name_list(self) -> list[str] | SyftError:
        if self.parent_job_id:
            parent = self.parent
            if isinstance(parent, SyftError):
                return parent
            parent_name_list = parent.ancestors_name_list
            if isinstance(parent_name_list, SyftError):
                return parent_name_list
            parent_name_list.append(parent.user_code_name)
            return parent_name_list
        return []

    def _repr_html_(self) -> str:
        # relative
        from ...util.notebook_ui.components.sync import CopyIDButton

        identifier = random.randint(1, 2**32)  # nosec
        result_tab_id = f"Result_{identifier}"
        logs_tab_id = f"Logs_{identifier}"
        job_type = "JOB" if not self.parent_job_id else "SUBJOB"
        ancestor_name_list = self.ancestors_name_list
        if isinstance(ancestor_name_list, SyftError):
            return ancestor_name_list
        api_header = f"{self.server_name}/jobs/" + "/".join(ancestor_name_list)
        copy_id_button = CopyIDButton(copy_text=str(self.id), max_width=60)
        button_html = copy_id_button.to_html()
        creation_time = self.creation_time[:-7] if self.creation_time else "--"
        updated_at = str(self.updated_at)[:-7] if self.updated_at else "--"

        user_repr = "--"
        if self.requested_by and not isinstance(
            requesting_user := self.requesting_user, SyftError
        ):
            user_repr = f"{requesting_user.name} {requesting_user.email}"

        worker_attr = ""
        if self.job_worker_id:
            worker = self.worker
            if not isinstance(worker, SyftError):
                worker_pool_id_button = CopyIDButton(
                    copy_text=str(worker.worker_pool_name), max_width=60
                )
                worker_attr = f"""
                    <div style="margin-top: 6px; margin-bottom: 6px;">
                    <span style="font-weight: 700; line-weight: 19.6px; font-size: 14px; font: 'Open Sans'">
                        Worker Pool:</span>
                        {worker.name} on worker {worker_pool_id_button.to_html()}
                    </div>
                """

        logs = self.logs(_print=False)
        logs_lines = logs.strip().split("\n") if logs else []
        logs_lines.insert(0, "<strong>Message</strong>")

        logs_lines = [f"<code>{line}</code>" for line in logs_lines]
        logs_lines_html = "\n".join(logs_lines)

        template = Template(job_repr_template)
        return template.substitute(
            job_type=job_type,
            api_header=api_header,
            user_code_name=self.user_code_name,
            button_html=button_html,
            status=self.status.value.title(),
            creation_time=creation_time,
            updated_at=updated_at,
            worker_attr=worker_attr,
            no_subjobs=len(self.subjobs),
            logs_tab_id=logs_tab_id,
            result_tab_id=result_tab_id,
            identifier=identifier,
            logs_lines_html=logs_lines_html,
            result=self.result,
            user_repr=user_repr,
        )

    def wait(
        self, job_only: bool = False, timeout: int | None = None
    ) -> Any | SyftNotReady | SyftError:
        self.fetch()
        if self.resolved:
            return self.resolve

        api = APIRegistry.api_for(
            server_uid=self.syft_server_location,
            user_verify_key=self.syft_client_verify_key,
        )

        if api is None:
            raise ValueError(
                f"Can't access Syft API. You must login to server with id '{self.syft_server_location}'"
            )

        workers = api.services.worker.get_all()
        if not isinstance(workers, SyftError) and len(workers) == 0:
            return SyftError(
                message=f"Server {self.syft_server_location} has no workers. "
                f"You need to start a worker to run jobs "
                f"by setting n_consumers > 0."
            )

        print_warning = True
        counter = 0
        while True:
            self.fetch()
            if self.resolved:
                if isinstance(self.result, SyftError | Err) or self.status in [  # type: ignore[unreachable]
                    JobStatus.ERRORED,
                    JobStatus.INTERRUPTED,
                ]:
                    return self.result
                break
            if print_warning and self.result is not None:
                result_obj = api.services.action.get(  # type: ignore[unreachable]
                    self.result.id, resolve_nested=False
                )
                if result_obj.is_link and job_only:
                    print(
                        "You're trying to wait on a job that has a link as a result."
                        "This means that the job may be ready but the linked result may not."
                        "Use job.wait().get() instead to wait for the linked result."
                    )
                    print_warning = False

            sleep(1)

            if timeout is not None:
                counter += 1
                if counter > timeout:
                    return SyftError(message="Reached Timeout!")

        # if self.resolve returns self.result as error, then we
        # return SyftError and not wait for the result
        # otherwise if a job is resolved and not errored out, we wait for the result
        if not job_only and self.result is not None:  # type: ignore[unreachable]
            self.result.wait(timeout)

        return self.resolve  # type: ignore[unreachable]

    @property
    def resolve(self) -> Any | SyftNotReady:
        if not self.resolved:
            self.fetch()

        if self.resolved:
            return self.result
        return SyftNotReady(message=f"{self.id} not ready yet.")

    def get_sync_dependencies(self, context: AuthedServiceContext) -> list[UID]:  # type: ignore
        dependencies = []
        if self.result is not None and isinstance(self.result, ActionObject):
            dependencies.append(self.result.id.id)

        if self.log_id:
            dependencies.append(self.log_id)

        subjobs = self.get_subjobs(context)
        if isinstance(subjobs, SyftError):
            return subjobs

        subjob_ids = [subjob.id for subjob in subjobs]
        dependencies.extend(subjob_ids)

        if self.user_code_id is not None:
            dependencies.append(self.user_code_id)

        output = context.server.get_service("outputservice").get_by_job_id(  # type: ignore
            context, self.id
        )
        if isinstance(output, SyftError):
            return output
        elif output is not None:
            dependencies.append(output.id)

        return dependencies


class JobInfo(SyftObject):
    __canonical_name__ = "JobInfo"
    __version__ = SYFT_OBJECT_VERSION_1

    __repr_attrs__ = [
        "resolved",
        "status",
        "n_iters",
        "current_iter",
        "creation_time",
    ]
    __public_metadata_attrs__ = [
        "resolved",
        "status",
        "n_iters",
        "current_iter",
        "creation_time",
    ]
    # Separate check if the job has logs, result, or metadata
    # None check is not enough because the values we set could be None
    includes_metadata: bool
    includes_result: bool
    # TODO add logs (error reporting PRD)

    user_code_id: UID | None = None
    resolved: bool | None = None
    status: JobStatus | None = None
    n_iters: int | None = None
    current_iter: int | None = None
    creation_time: str | None = None

    result: ActionObject | None = None

    def _repr_html_(self) -> str:
        metadata_str = ""
        if self.includes_metadata:
            metadata_str += "<h4>Public metadata</h4>"
            for attr in self.__public_metadata_attrs__:
                value = getattr(self, attr, None)
                if value is not None:
                    metadata_str += f"<p style='margin-left: 10px;'><strong>{attr}:</strong> {value}</p>"

        result_str = "<h4>Result</h4>"
        if self.includes_result:
            result_str += f"<p style='margin-left: 10px;'>{str(self.result)}</p>"
        else:
            result_str += "<p style='margin-left: 10px;'><i>No result included</i></p>"

        return f"""
            <style>
            .job-info {{color: {SURFACE[options.color_theme]};}}
            </style>
            <div class='job-info'>
                <h3>JobInfo</h3>
                {metadata_str}
                {result_str}
            </div>
        """

    @classmethod
    def from_job(
        cls,
        job: Job,
        metadata: bool = False,
        result: bool = False,
    ) -> Self:
        info = cls(
            includes_metadata=metadata,
            includes_result=result,
            user_code_id=job.user_code_id,
        )

        if metadata:
            for attr in cls.__public_metadata_attrs__:
                setattr(info, attr, getattr(job, attr))

        if result:
            if not job.resolved:
                raise ValueError("Cannot sync result of unresolved job")
            if not isinstance(job.result, ActionObject):
                raise ValueError("Could not sync result of job")
            info.result = job.result

        return info


@instrument
@serializable(canonical_name="JobStash", version=1)
class JobStash(BaseUIDStoreStash):
    object_type = Job
    settings: PartitionSettings = PartitionSettings(
        name=Job.__canonical_name__, object_type=Job
    )

    def __init__(self, store: DocumentStore) -> None:
        super().__init__(store=store)

    def set_result(
        self,
        credentials: SyftVerifyKey,
        item: Job,
        add_permissions: list[ActionObjectPermission] | None = None,
    ) -> Result[Job | None, str]:
        valid = self.check_type(item, self.object_type)
        if valid.is_err():
            return SyftError(message=valid.err())
        return super().update(credentials, item, add_permissions)

    def get_by_result_id(
        self,
        credentials: SyftVerifyKey,
        res_id: UID,
    ) -> Result[Job | None, str]:
        res = self.get_all(credentials)
        if res.is_err():
            return res
        else:
            res = res.ok()
            # beautiful query
            res = [
                x
                for x in res
                if isinstance(x.result, ActionObject) and x.result.id.id == res_id
            ]
            if len(res) == 0:
                return Ok(None)
            elif len(res) > 1:
                return Err("multiple Jobs found")
            else:
                return Ok(res[0])

    def get_by_parent_id(
        self, credentials: SyftVerifyKey, uid: UID
    ) -> Result[Job | None, str]:
        qks = QueryKeys(
            qks=[PartitionKey(key="parent_job_id", type_=UID).with_obj(uid)]
        )
        item = self.query_all(credentials=credentials, qks=qks)
        return item

    def delete_by_uid(
        self, credentials: SyftVerifyKey, uid: UID
    ) -> Result[SyftSuccess, str]:
        qk = UIDPartitionKey.with_obj(uid)
        result = super().delete(credentials=credentials, qk=qk)
        if result.is_ok():
            return Ok(SyftSuccess(message=f"ID: {uid} deleted"))
        return result

    def get_active(self, credentials: SyftVerifyKey) -> Result[SyftSuccess, str]:
        qks = QueryKeys(
            qks=[
                PartitionKey(key="status", type_=JobStatus).with_obj(
                    JobStatus.PROCESSING
                )
            ]
        )
        return self.query_all(credentials=credentials, qks=qks)

    def get_by_worker(
        self, credentials: SyftVerifyKey, worker_id: str
    ) -> Result[list[Job], str]:
        qks = QueryKeys(
            qks=[PartitionKey(key="job_worker_id", type_=str).with_obj(worker_id)]
        )
        return self.query_all(credentials=credentials, qks=qks)

    def get_by_user_code_id(
        self, credentials: SyftVerifyKey, user_code_id: UID
    ) -> Result[list[Job], str]:
        qks = QueryKeys(
            qks=[PartitionKey(key="user_code_id", type_=UID).with_obj(user_code_id)]
        )

        return self.query_all(credentials=credentials, qks=qks)
